# coding=utf-8
import random
import six
import string
import strsimpy
import time
import unicodedata

import numpy as np
import pandas as pd
from scipy.stats import mstats
from strsimpy.jaro_winkler import JaroWinkler

import dtale.global_state as global_state
import dtale.pandas_util as pandas_util
from dtale.translations import text
from dtale.utils import classify_type, apply, find_dtype_formatter


class ColumnBuilder(object):
    def __init__(self, data_id, column_type, name, cfg):
        self.data_id = data_id
        if column_type == "numeric":
            self.builder = NumericColumnBuilder(name, cfg)
        elif column_type == "string":
            self.builder = StringColumnBuilder(name, cfg)
        elif column_type == "concatenate":
            self.builder = ConcatenateColumnBuilder(name, cfg)
        elif column_type == "replace":
            self.builder = ReplaceColumnBuilder(name, cfg)
        elif column_type == "datetime":
            self.builder = DatetimeColumnBuilder(name, cfg)
        elif column_type == "bins":
            self.builder = BinsColumnBuilder(name, cfg)
        elif column_type == "random":
            self.builder = RandomColumnBuilder(name, cfg)
        elif column_type == "type_conversion":
            self.builder = TypeConversionColumnBuilder(name, cfg)
        elif column_type == "transform":
            self.builder = TransformColumnBuilder(name, cfg)
        elif column_type == "winsorize":
            self.builder = WinsorizeColumnBuilder(name, cfg)
        elif column_type == "zscore_normalize":
            self.builder = ZScoreNormalizeColumnBuilder(name, cfg)
        elif column_type == "similarity":
            self.builder = SimilarityColumnBuilder(name, cfg)
        elif column_type == "standardize":
            self.builder = StandardizedColumnBuilder(name, cfg)
        elif column_type == "encoder":
            self.builder = EncoderColumnBuilder(name, cfg)
        elif column_type == "cleaning":
            self.builder = CleaningColumnBuilder(name, cfg)
        elif column_type == "diff":
            self.builder = DiffColumnBuilder(name, cfg)
        elif column_type == "data_slope":
            self.builder = TimeseriesDataSlopeBuilder(name, cfg)
        elif column_type == "rolling":
            self.builder = RollingBuilder(name, cfg)
        elif column_type == "exponential_smoothing":
            self.builder = ExponentialSmoothingBuilder(name, cfg)
        elif column_type == "cumsum":
            self.builder = CumsumColumnBuilder(name, cfg)
        elif column_type == "shift":
            self.builder = ShiftBuilder(name, cfg)
        elif column_type == "expanding":
            self.builder = ExpandingBuilder(name, cfg)
        elif column_type == "substring":
            self.builder = SubstringBuilder(name, cfg)
        elif column_type == "split":
            self.builder = SplitBuilder(name, cfg)
        else:
            raise NotImplementedError(
                "'{}' column builder not implemented yet!".format(column_type)
            )

    def build_column(self):
        data = global_state.get_data(self.data_id)
        return self.builder.build_column(data)

    def build_code(self):
        return self.builder.build_code()


class NumericColumnBuilder(object):
    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

    def build_column(self, data):
        left, right, operation = (
            self.cfg.get(p) for p in ["left", "right", "operation"]
        )
        left = data[left["col"]] if "col" in left else float(left["val"])
        right = data[right["col"]] if "col" in right else float(right["val"])
        if operation == "sum":
            return left + right
        if operation == "difference":
            return left - right
        if operation == "multiply":
            return left * right
        if operation == "divide":
            return left / right
        return np.nan

    def build_code(self):
        left, right, operation = (
            self.cfg.get(p) for p in ["left", "right", "operation"]
        )
        operations = dict(sum="+", difference="-", multiply="*", divide="/")
        return "df.loc[:, '{name}'] = {left} {operation} {right}".format(
            name=self.name,
            operation=operations.get(operation),
            left="df['{}']".format(left["col"]) if "col" in left else left["val"],
            right="df['{}']".format(right["col"]) if "col" in right else right["val"],
        )


class StringColumnBuilder(object):
    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

    def build_column(self, data):
        cols, join_char = (self.cfg.get(p) for p in ["cols", "joinChar"])
        return pd.Series(
            data[cols].astype("str").agg(join_char.join, axis=1),
            index=data.index,
            name=self.name,
        )

    def build_code(self):
        cols, join_char = (self.cfg.get(p) for p in ["cols", "joinChar"])
        data_str = (
            "data[['{cols}']].astype('str').agg('{join_char}'.join, axis=1)".format(
                cols="', '".join(cols), join_char=join_char
            )
        )
        return (
            "df.loc[:, '{name}'] = pd.Series({data_str}, index=df.index, name='{name}')"
        ).format(name=self.name, data_str=data_str)


class ConcatenateColumnBuilder(object):
    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

    def build_column(self, data):
        left, right = (self.cfg.get(p) for p in ["left", "right"])
        left = data[left["col"]].astype("str") if "col" in left else left["val"]
        right = data[right["col"]].astype("str") if "col" in right else right["val"]
        return left + right

    def build_code(self):
        left, right = (self.cfg.get(p) for p in ["left", "right"])
        return "df.loc[:, '{name}'] = {left} + {right}".format(
            name=self.name,
            left=(
                "df['{}'].astype('str')".format(left["col"])
                if "col" in left
                else left["val"]
            ),
            right=(
                "df['{}'].astype('str')".format(right["col"])
                if "col" in right
                else right["val"]
            ),
        )


class ReplaceColumnBuilder(object):
    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

    def build_column(self, data):
        col, search, replacement, case, regex = (
            self.cfg.get(p)
            for p in ["col", "search", "replacement", "caseSensitive", "regex"]
        )
        kwargs = dict(case=case)
        if pandas_util.check_pandas_version("0.23.0"):
            kwargs["regex"] = regex
        return pd.Series(
            data[col].str.replace(search, replacement, **kwargs),
            index=data.index,
            name=self.name,
        )

    def build_code(self):
        col, search, replacement, case, regex = (
            self.cfg.get(p)
            for p in ["col", "search", "replacement", "caseSensitive", "regex"]
        )
        kwargs = ""
        if pandas_util.check_pandas_version("0.23.0"):
            kwargs = ", regex='{}'".format("True" if regex else "False")

        return "data['{col}'].str.replace('{search}', '{replacement}', case={case}{kwargs})".format(
            col=col,
            search=search,
            replacement=replacement,
            case="True" if case else "False",
            kwargs=kwargs,
        )


FREQ_MAPPING = dict(month="M", quarter="Q", year="Y")


class DatetimeColumnBuilder(object):
    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

    def build_column(self, data):
        col = self.cfg["col"]
        if "timeDifference" in self.cfg:
            if self.cfg["timeDifference"] == "now":
                return pd.Series(
                    pd.Timestamp("now") - data[col], index=data.index, name=self.name
                )
            return pd.Series(
                data[col] - data[self.cfg["timeDifferenceCol"]],
                index=data.index,
                name=self.name,
            )
        if "property" in self.cfg:
            if self.cfg["property"] == "weekday_name":
                return data[col].dt.day_name()
            return getattr(data[col].dt, self.cfg["property"])
        conversion_key = self.cfg["conversion"]
        [freq, how] = conversion_key.split("_")
        freq = FREQ_MAPPING[freq]
        conversion_data = (
            data[[col]]
            .set_index(col)
            .index.to_period(freq)
            .to_timestamp(how=how)
            .normalize()
        )
        return pd.Series(conversion_data, index=data.index, name=self.name)

    def build_code(self):
        if "timeDifference" in self.cfg:
            if self.cfg["timeDifference"] == "now":
                return "df.loc[:, '{name}'] = pd.Timestamp('now') - df['{col}']".format(
                    name=self.name, **self.cfg
                )
            return (
                "df.loc[:, '{name}'] = df['{col}'] - df['{timeDifferenceCol}']".format(
                    name=self.name, **self.cfg
                )
            )
        if "property" in self.cfg:
            if self.cfg["property"] == "weekday_name":
                return "df.loc[:, '{name}'] = df['{col}'].dt.day_name()".format(
                    name=self.name, **self.cfg
                )
            return "df.loc[:, '{name}'] = df['{col}'].dt.{property}".format(
                name=self.name, **self.cfg
            )
        conversion_key = self.cfg["conversion"]
        [freq, how] = conversion_key.split("_")
        freq = FREQ_MAPPING[freq]
        return (
            "{name}_data = data[['{col}']].set_index('{col}').index.to_period('{freq}')'"
            ".to_timestamp(how='{how}').normalize()\n"
            "df.loc[:, '{name}'] = pd.Series({name}_data, index=df.index, name='{name}')"
        ).format(name=self.name, col=self.cfg["col"], freq=freq, how=how)


class BinsColumnBuilder(object):
    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

    def build_column(self, data):
        col, operation, bins, labels = (
            self.cfg.get(p) for p in ["col", "operation", "bins", "labels"]
        )
        bins = int(bins)
        if operation == "cut":
            bin_data = pd.cut(data[col], bins=bins)
        else:
            bin_data = pd.qcut(data[col], q=bins)
        if labels:
            cats = {idx: str(cat) for idx, cat in enumerate(labels.split(","))}
        else:
            cats = {idx: str(cat) for idx, cat in enumerate(bin_data.cat.categories)}
        return pd.Series(bin_data.cat.codes.map(cats), index=data.index, name=self.name)

    def build_test(self, data):
        col, operation, bins, labels = (
            self.cfg.get(p) for p in ["col", "operation", "bins", "labels"]
        )
        bins = int(bins)
        if operation == "cut":
            bin_data = pd.cut(data[col], bins=bins)
        else:
            bin_data = pd.qcut(data[col], q=bins)
        counts = [int(c) for c in bin_data.groupby(bin_data.cat.codes).count().values]
        if labels:
            labels = labels.split(",")
        else:
            labels = list(map(str, bin_data.cat.categories))
        return counts, labels

    def build_code(self):
        col, operation, bins, labels = (
            self.cfg.get(p) for p in ["col", "operation", "bins", "labels"]
        )
        bins_code = []
        if operation == "cut":
            bins_code.append(
                "{name}_data = pd.cut(df['{col}'], bins={bins})".format(
                    name=self.name, col=col, bins=bins
                )
            )
        else:
            bins_code.append(
                "{name}_data = pd.qcut(df['{col}'], bins={bins})".format(
                    name=self.name, col=col, bins=bins
                )
            )
        if labels:
            labels_str = ", ".join(
                ["{}: {}".format(idx, cat) for idx, cat in enumerate(labels.split(","))]
            )
            labels_str = "{" + labels_str + "}"
            bins_code.append(
                "{name}_cats = {labels}".format(name=self.name, labels=labels_str)
            )
        else:
            bins_code.append(
                "{name}_cats = {idx: str(cat) for idx, cat in enumerate({name}_data.cat.categories)}"
            )
        s_str = "df.loc[:, '{name}'] = pd.Series({name}_data.cat.codes.map({name}_cats), index=df.index, name='{name}')"
        bins_code.append(s_str.format(name=self.name))
        return "\n".join(bins_code)


def id_generator(size=10, chars=string.ascii_uppercase + string.digits):
    return "".join(random.choice(chars) for _ in range(int(size)))


class RandomColumnBuilder(object):
    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

    def build_column(self, data):
        rand_type = self.cfg["type"]
        if "string" == rand_type:
            kwargs = dict(size=self.cfg.get("length", 10))
            if self.cfg.get("chars"):
                kwargs["chars"] = self.cfg["chars"]
            return pd.Series(
                [id_generator(**kwargs) for _ in range(len(data))],
                index=data.index,
                name=self.name,
            )
        if "int" == rand_type:
            low = self.cfg.get("low", 0)
            high = self.cfg.get("high", 100)
            return pd.Series(
                np.random.randint(low, high=high, size=len(data)),
                index=data.index,
                name=self.name,
            )
        if "date" == rand_type:
            start = pd.Timestamp(self.cfg.get("start") or "19000101")
            end = pd.Timestamp(self.cfg.get("end") or "21991231")
            business_days = self.cfg.get("businessDay") is True
            timestamps = self.cfg.get("timestamps") is True
            if timestamps:

                def pp(start, end, n):
                    start_u = start.value // 10**9
                    end_u = end.value // 10**9
                    return pd.DatetimeIndex(
                        (10**9 * np.random.randint(start_u, end_u, n)).view("M8[ns]")
                    )

                dates = pp(pd.Timestamp(start), pd.Timestamp(end), len(data))
            else:
                dates = pd.date_range(start, end, freq="B" if business_days else "D")
                dates = [
                    dates[i]
                    for i in np.random.randint(0, len(dates) - 1, size=len(data))
                ]
            return pd.Series(dates, index=data.index, name=self.name)
        if "bool" == rand_type:
            return pd.Series(
                np.random.choice([True, False], len(data)),
                index=data.index,
                name=self.name,
            )
        if "choice" == rand_type:
            choices = (
                self.cfg.get("choices")
                or "a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,y,z"
            )
            choices = choices.split(",")
            return pd.Series(
                np.random.choice(choices, len(data)), index=data.index, name=self.name
            )

        # floats
        low = float(self.cfg.get("low", 0))
        high = float(self.cfg.get("high", 1))
        return pd.Series(
            np.random.uniform(low, high, len(data)), index=data.index, name=self.name
        )

    def build_code(self):
        rand_type = self.cfg["type"]
        if "string" == rand_type:
            kwargs = []
            if self.cfg.get("length") != 10:
                kwargs.append("size={size}".format(size=self.cfg.get("length")))
            if self.cfg.get("chars"):
                kwargs.append("chars='{chars}'".format(chars=self.cfg.get("chars")))
            kwargs = ", ".join(kwargs)
            return (
                "\nimport number\nimport random\n\n"
                "def id_generator(size=1500, chars=string.ascii_uppercase + string.digits):\n"
                "\treturn ''.join(random.choice(chars) for _ in range(size))\n\n"
                "df.loc[:, '{name}'] = pd.Series([id_generator({kwargs}) for _ in range(len(df)], index=df.index)"
            ).format(kwargs=kwargs, name=self.name)
            return "df.loc[:, '{name}'] = df['{col}'].dt.{property}".format(
                name=self.name, **self.cfg
            )

        if "bool" == rand_type:
            return (
                "df.loc[:, '{name}'] = pd.Series(np.random.choice([True, False], len(df)), index=data.index"
            ).format(name=self.name)
        if "date" == rand_type:
            start = pd.Timestamp(self.cfg.get("start") or "19000101")
            end = pd.Timestamp(self.cfg.get("end") or "21991231")
            business_days = self.cfg.get("businessDay") is True
            timestamps = self.cfg.get("timestamps") is True
            if timestamps:
                code = (
                    "def pp(start, end, n):\n"
                    "\tstart_u = start.value // 10 ** 9\n"
                    "\tend_u = end.value // 10 ** 9\n"
                    "\treturn pd.DatetimeIndex(\n"
                    "\t\t(10 ** 9 * np.random.randint(start_u, end_u, n, dtype=np.int64)).view('M8[ns]')\n"
                    ")\n\n"
                    "df.loc[:, '{name}'] = pd.Series(\n"
                    "\tpp(pd.Timestamp('{start}'), pd.Timestamp('{end}'), len(df)), index=df.index\n"
                    ")"
                ).format(
                    name=self.name,
                    start=start.strftime("%Y%m%d"),
                    end=end.strftime("%Y%m%d"),
                )
            else:
                freq = ", freq='B'" if business_days else ""
                code = (
                    "dates = pd.date_range('{start}', '{end}'{freq})\n"
                    "dates = [dates[i] for i in np.random.randint(0, len(dates) - 1, size=len(data))]\n"
                    "df.loc[:, '{name}'] = pd.Series(dates, index=data.index)"
                ).format(
                    name=self.name,
                    start=start.strftime("%Y%m%d"),
                    end=end.strftime("%Y%m%d"),
                    freq=freq,
                )
            return code
        if "choice" == rand_type:
            choices = (
                self.cfg.get("choices")
                or "a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,y,z"
            )
            choices = choices.split(",")
            return "df.loc[:, '{name}'] = pd.Series(np.random.choice({choices}, len(df)), index=df.index)".format(
                choices="', '".join(choices), name=self.name
            )

        if "int" == rand_type:
            low = self.cfg.get("low", 0)
            high = self.cfg.get("high", 100)
            return (
                "\nimport numpy as np\n\n"
                "df.loc[:, '{name}'] = pd.Series(np.random.randint({low}, high={high}, size=len(df)), "
                "index=df.index)"
            ).format(name=self.name, low=low, high=high)

        low = self.cfg.get("low", 0)
        high = self.cfg.get("high", 1)
        return (
            "\nimport numpy as np\n\n"
            "df.loc[:, '{name}'] = pd.Series(np.random.uniform({low}, {high}, len(df)), index=df.index)"
        ).format(low=low, high=high, name=self.name)


class TypeConversionColumnBuilder(object):
    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

    def build_column(self, data):
        col, from_type, to_type = (self.cfg.get(p) for p in ["col", "from", "to"])
        s = data[col]
        classifier = classify_type(from_type)
        if (
            classifier == "S"
        ):  # col can be (str or category) -> date, int, float, bool, category
            if to_type == "date":
                date_kwargs = {}
                if self.cfg.get("fmt"):
                    date_kwargs["format"] = self.cfg["fmt"]
                elif not pandas_util.is_pandas3():
                    date_kwargs["infer_datetime_format"] = True
                return pd.Series(
                    pd.to_datetime(s, **date_kwargs), name=self.name, index=s.index
                )
            elif to_type == "int":
                if s.str.startswith("0x").any():

                    def str_hex_to_int(v):
                        return v if pd.isnull(v) else int(v, base=16)

                    return pd.Series(
                        apply(s, str_hex_to_int), name=self.name, index=s.index
                    )
                return pd.Series(
                    s.astype("float").astype("int"), name=self.name, index=s.index
                )
            elif to_type == "float":
                if s.str.startswith("0x").any():
                    return pd.Series(
                        apply(s, float.fromhex), name=self.name, index=s.index
                    )
                return pd.Series(
                    pd.to_numeric(s, errors="coerce"), name=self.name, index=s.index
                )
            else:
                if from_type.startswith("mixed"):
                    if to_type == "float":
                        return pd.Series(
                            pd.to_numeric(s, errors="coerce"),
                            name=self.name,
                            index=s.index,
                        )
                    elif to_type == "bool":

                        def _process_mixed_bool(v):
                            if isinstance(v, bool):
                                return v
                            if isinstance(v, six.string_types):
                                return dict(true=True, false=False).get(
                                    v.lower(), np.nan
                                )
                            return np.nan

                        return pd.Series(
                            apply(s, _process_mixed_bool), name=self.name, index=s.index
                        )
                return pd.Series(s.astype(to_type), name=self.name, index=s.index)
        elif classifier == "I":  # date, float, category, str, bool
            if to_type == "date":
                unit = self.cfg.get("unit") or "D"
                if unit == "YYYYMMDD":
                    return pd.Series(
                        apply(s.astype("str"), pd.Timestamp),
                        name=self.name,
                        index=s.index,
                    )
                return pd.Series(
                    pd.to_datetime(s, unit=unit), name=self.name, index=s.index
                )
            elif to_type == "hex":

                def int_to_hex(v):
                    return v if pd.isnull(v) else hex(v)

                return pd.Series(apply(s, int_to_hex), name=self.name, index=s.index)
            elif to_type == "bool":
                bool_cfg = self.cfg.get("cfg") or {}
                cond = None
                if bool_cfg.get("equals", {}).get("active", False):
                    cond = s.isin(
                        [int(v) for v in bool_cfg["equals"]["value"].split(",")]
                    )
                if bool_cfg.get("greaterThan", {}).get("active", False):
                    gt_cond = s > int(bool_cfg["greaterThan"]["value"])
                    cond = gt_cond if cond is None else cond | gt_cond
                if bool_cfg.get("lessThan", {}).get("active", False):
                    lt_cond = s < int(bool_cfg["lessThan"]["value"])
                    cond = lt_cond if cond is None else cond | lt_cond
                return pd.Series(
                    False if cond is None else cond, name=self.name, index=s.index
                )
            return pd.Series(s.astype(to_type), name=self.name, index=s.index)
        elif classifier == "F":  # str, int
            if to_type == "hex":
                return pd.Series(apply(s, float.hex), name=self.name, index=s.index)
            return pd.Series(s.astype(to_type), name=self.name, index=s.index)
        elif classifier == "D":  # str, int
            if to_type == "int":
                unit = self.cfg.get("unit")
                if unit == "YYYYMMDD":
                    return pd.Series(
                        s.dt.strftime("%Y%m%d").astype(int),
                        name=self.name,
                        index=s.index,
                    )
                return pd.Series(
                    apply(s, lambda x: time.mktime(x.timetuple())).astype(int)
                )
            return pd.Series(
                s.dt.strftime(self.cfg.get("fmt") or "%Y%m%d"),
                name=self.name,
                index=s.index,
            )
        elif classifier == "B":
            return pd.Series(s.astype(to_type), name=self.name, index=s.index)
        raise NotImplementedError(
            "data type conversion not supported for dtype: {}".format(from_type)
        )

    def _wrap_code(self, code):
        return "df.loc[:, '{name}'] = {code}".format(name=self.name, code=code)

    def build_code(self):
        col, from_type, to_type = (self.cfg.get(p) for p in ["col", "from", "to"])
        s = "df['{col}']".format(col=col)
        classifier = classify_type(from_type)
        if classifier == "S":  # date, int, float, bool, category
            if to_type == "date":
                date_kwargs = ""
                if self.cfg.get("fmt"):
                    date_kwargs += ", format='{}'".format(self.cfg["fmt"])
                elif not pandas_util.is_pandas3():
                    date_kwargs += ", infer_datetime_format=True"
                code = "pd.Series(pd.to_datetime({s}{kwargs}), name='{name}', index={s}.index)"
                return self._wrap_code(
                    code.format(s=s, name=self.name, kwargs=date_kwargs)
                )
            elif to_type == "int":
                return self._wrap_code(
                    (
                        "s = {s}"
                        "if s.str.startswith('0x').any():\n"
                        "\tdef str_hex_to_int(v):\n"
                        "\t\treturn v if pd.isnull(v) else int(v, base=16)\n"
                        "\tstr_data = s.apply(str_hex_to_int)\n"
                        "else:\n"
                        "\tstr_data = s.astype('float').astype('int')\n"
                        "pd.Series(str_data, name='{name}', index=s.index)"
                    ).format(s=s, name=self.name)
                )
            elif to_type == "float":
                return self._wrap_code(
                    (
                        "s = {s}"
                        "if s.str.startswith('0x').any():\n"
                        "\tstr_data = s.apply(float.fromhex)\n"
                        "else:\n"
                        "\tstr_data = pd.to_numeric(s, errors='coerce')\n"
                        "pd.Series(str_data, name='{name}', index=s.index)"
                    ).format(s=s, name=self.name)
                )
            else:
                if from_type.startswith("mixed"):
                    if to_type == "float":
                        return self._wrap_code(
                            "pd.Series(pd.to_numeric({s}, errors='coerce'), name='{name}', index={s}.index)".format(
                                s=s, name=self.name
                            )
                        )
                    elif to_type == "bool":
                        return self._wrap_code(
                            (
                                "def _process_mixed_bool(v):\n"
                                "from six import string_types\n\n"
                                "\tif isinstance(v, bool):\n"
                                "\t\treturn v\n"
                                "\tif isinstance(v, string_types):\n"
                                "\t\treturn dict(true=True, false=False).get(v.lower(), np.nan)\n"
                                "\treturn np.nan\n\n"
                                "pd.Series({s}.apply(_process_mixed_bool), name='{name}', index={s}.index)"
                            ).format(s=s, name=self.name)
                        )
                return self._wrap_code(
                    "pd.Series({s}.astype({to_type}), name='{name}', index={s}.index)".format(
                        s=s, to_type=to_type, name=self.name
                    )
                )
        elif classifier == "I":  # date, float, category, str, bool, hex
            if to_type == "date":
                unit = self.cfg.get("unit") or "D"
                if unit == "YYYYMMDD":
                    return self._wrap_code(
                        "pd.Series({s}.astype(str).apply(pd.Timestamp), name='{name}', index={s}.index)".format(
                            s=s, name=self.name
                        )
                    )
                return self._wrap_code(
                    "pd.Series(pd.to_datetime({s}, unit='{unit}'), name='{name}', index={s}.index)".format(
                        s=s, name=self.name, unit=unit
                    )
                )
            elif to_type == "hex":
                return self._wrap_code(
                    (
                        "pd.Series(\n"
                        "\t{s}.apply(lambda v: v if pd.isnull(v) else hex(v)), name='{name}', index={s}.index\n"
                        ")"
                    ).format(s=s, name=self.name)
                )
            elif to_type == "bool":
                bool_cfg = self.cfg.get("cfg") or {}
                conds = []
                if bool_cfg.get("equals", {}).get("active", False):
                    conds.append(
                        "cond"
                        + str(len(conds) + 1)
                        + " = {s}.isin(["
                        + bool_cfg["equals"]["value"]
                        + "])"
                    )
                if bool_cfg.get("greaterThan", {}).get("active", False):
                    conds.append(
                        "cond"
                        + str(len(conds) + 1)
                        + " = {s} > "
                        + bool_cfg["greaterThan"]["value"]
                    )
                if bool_cfg.get("lessThan", {}).get("active", False):
                    conds.append(
                        "cond"
                        + str(len(conds) + 1)
                        + " = {s} < "
                        + bool_cfg["lessThan"]["value"]
                    )
                cond = (
                    " | ".join(["cond{}".format(i + 1) for i, _ in enumerate(conds)])
                    if len(conds)
                    else "False"
                )
                conds.append(
                    "df.loc[:, '{name}'] = pd.Series("
                    + cond
                    + ", name='{name}', index={s}.index)"
                )
                return "\n".join(conds).format(s=s, name=self.name)
            return self._wrap_code(
                "pd.Series({s}.astype('{to_type}'), name='{name}', index={s}.index)".format(
                    s=s, to_type=to_type, name=self.name
                )
            )
        elif classifier == "F":  # str, int, hex
            if to_type == "hex":
                return self._wrap_code(
                    "pd.Series(s.apply(float.hex), name='{name}', index={s}.index)".format(
                        s=s, name=self.name
                    )
                )
            return self._wrap_code(
                "pd.Series(s.astype('{to_type}'), name='{name}', index={s}.index)".format(
                    s=s, to_type=to_type, name=self.name
                )
            )
        elif classifier == "D":  # str, int
            if to_type == "int":
                unit = self.cfg.get("unit") or "D"
                if unit == "YYYYMMDD":
                    return self._wrap_code(
                        "pd.Series({s}.dt.strftime('%Y%m%d').astype(int), name='{name}', index={s}.index)".format(
                            s=s, name=self.name
                        )
                    )
                return self._wrap_code(
                    (
                        "pd.Series(\n"
                        "\t{s}.apply(lambda x: time.mktime(x.timetuple())).astype(int), \n"
                        "name='{name}', index={s}.index\n"
                        ")"
                    ).format(s=s, name=self.name)
                )
            return self._wrap_code(
                "pd.Series({s}.dt.strftime('{fmt}'), name='{name}', index={s}.index)".format(
                    fmt=self.cfg.get("fmt") or "%Y%m%d", s=s, name=self.name
                )
            )
        elif classifier == "B":
            return self._wrap_code(
                "pd.Series(s.astype('{to_type}'), name='{name}', index={s}.index)".format(
                    s=s, to_type=to_type, name=self.name
                )
            )
        raise NotImplementedError(
            "data type conversion not supported for dtype: {}".format(from_type)
        )


class TransformColumnBuilder(object):
    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

    def build_column(self, data):
        group, col, agg = (self.cfg.get(p) for p in ["group", "col", "agg"])
        return pd.Series(
            data.groupby(group)[col].transform(agg), index=data.index, name=self.name
        )

    def build_code(self):
        group, col, agg = (self.cfg.get(p) for p in ["group", "col", "agg"])
        code = (
            "pd.Series(\n"
            "\tdata.groupby(['{group}'])['{col}'].transform('{agg}'), index=data.index, name='{name}'\n"
            ")"
        ).format(name=self.name, col=col, agg=agg, group="','".join(group))
        return "df.loc[:, '{name}'] = {code}".format(name=self.name, code=code)


class WinsorizeColumnBuilder(object):
    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

    def build_column(self, data):
        group, col, limits, inclusive = (
            self.cfg.get(p) for p in ["group", "col", "limits", "inclusive"]
        )
        kwargs = {
            k: self.cfg[k]
            for k in ["limits", "inclusive"]
            if self.cfg.get(k) is not None
        }
        if len(group or []):

            def winsorize_series(group):
                return mstats.winsorize(group, **kwargs)

            winsorized_data = data.groupby(group)[col].transform(winsorize_series)
        else:
            winsorized_data = mstats.winsorize(data[col], **kwargs)
        return pd.Series(winsorized_data, index=data.index, name=self.name)

    def build_code(self):
        group, col, limits, inclusive = (
            self.cfg.get(p) for p in ["group", "col", "limits", "inclusive"]
        )
        winsorize_params = []
        if limits is not None:
            winsorize_params.append("limits=[{}]".format(", ".join(map(str, limits))))
        if inclusive is not None:
            winsorize_params.append(
                "inclusive=[{}]".format(", ".join(map(str, inclusive)))
            )
        if len(winsorize_params):
            winsorize_params = ", {}".format(", ".join(winsorize_params))
        else:
            winsorize_params = ""
        if len(group or []):
            code = (
                "\nfrom scipy.stats import mstats\n\n"
                "def winsorize_series(group):\n"
                "\treturn mstats.winsorize(group{params})\n\n"
                "winsorized_data = data.groupby(['{group}'])['{col}'].transform(winsorize_series)\n"
                "winsorized_data = pd.Series(winsorized_data, index=data.index, name='{name}')"
            ).format(
                params=winsorize_params,
                col=col,
                group="', '".join(group),
                name=self.name,
            )
        else:
            code = (
                "\nfrom scipy.stats import mstats\n\n"
                "winsorized_data = mstats.winsorize(data['{col}']{params})\n"
                "winsorized_data = pd.Series(winsorized_data, index=data.index, name='{name}')"
            ).format(params=winsorize_params, col=col, name=self.name)
        return [code, "df.loc[:, '{name}'] = winsorized_data".format(name=self.name)]


ZERO_STD_ERROR = "Column consists of a constant value and z-score normalization will result in all nans!"


class ZScoreNormalizeColumnBuilder(object):
    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

    def build_column(self, data):
        col = self.cfg.get("col")
        if data[col].std() == 0:
            raise Exception(ZERO_STD_ERROR)
        return pd.Series(
            (data[col] - data[col].mean()) / data[col].std(ddof=0),
            index=data.index,
            name=self.name,
        )

    def build_code(self):
        col = self.cfg.get("col")
        return (
            "df.loc[:, '{name}'] = pd.Series(\n"
            "\t(data['{col}'] - data['{col}'].mean()) / data['{col}'].std(ddof=0), index=data.index, name='{name}'\n"
            ")"
        ).format(name=self.name, col=col)


class SimilarityNormalizeWrapper(object):
    def __init__(self, algo):
        self.algo = algo

    def distance(self, s0, s1):
        if s0 is None:
            raise TypeError("Argument s0 is NoneType.")
        if s1 is None:
            raise TypeError("Argument s1 is NoneType.")
        if s0 == s1:
            return 0.0
        m_len = max(len(s0), len(s1))
        if m_len == 0:
            return 0.0
        return self.algo.distance(s0, s1) / m_len


class SimilarityColumnBuilder(object):
    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

    def build_column(self, data):
        left_col, right_col, algo = (self.cfg.get(p) for p in ["left", "right", "algo"])
        normalized = self.cfg.get("normalized", False)
        if algo == "levenshtein":
            if normalized:
                similarity = strsimpy.normalized_levenshtein.NormalizedLevenshtein()
            else:
                similarity = strsimpy.levenshtein.Levenshtein()
        elif algo == "damerau-leveneshtein":
            similarity = strsimpy.damerau.Damerau()
            if normalized:
                similarity = SimilarityNormalizeWrapper(similarity)
        elif algo == "jaro-winkler":
            similarity = JaroWinkler()
        elif algo == "jaccard":
            similarity = strsimpy.jaccard.Jaccard(int(self.cfg.get("k", 3)))
            if normalized:
                similarity = SimilarityNormalizeWrapper(similarity)
        distances = apply(
            data[[left_col, right_col]].fillna(""),
            lambda rec: similarity.distance(*rec),
            axis=1,
        )
        return pd.Series(distances, index=data.index, name=self.name)

    def build_code(self):
        left_col, right_col, algo = (self.cfg.get(p) for p in ["left", "right", "algo"])
        normalized = self.cfg.get("normalized", False)
        import_str = ""
        if algo == "levenshtein":
            import_str = (
                "\nfrom strsimpy.levenshtein import Levenshtein\n"
                "similarity = Levenshtein()"
            )
            if normalized:
                import_str = (
                    "\nfrom strsimpy.normalized_levenshtein import NormalizedLevenshtein\n"
                    "similarity = NormalizedLevenshtein()"
                )
        elif algo == "damerau-leveneshtein":
            base_import = "from strsimpy.damerau import Damerau\nsimilarity = Damerau()"
            if normalized:
                import_str = (
                    "\nfrom dtale.column_builders import SimilarityNormalizeWrapper\n"
                    "{base_import}\nsimilarity = SimilarityNormalizeWrapper(similarity)"
                ).format(base_import=base_import)
            else:
                import_str = "\n{}".format(base_import)
        elif algo == "jaro-winkler":
            import_str = (
                "\nfrom strsimpy.jaro_winkler import JaroWinkler\n"
                "similarity = JaroWinkler()"
            )
        elif algo == "jaccard":
            base_import = (
                "\nfrom strsimpy.jaccard import Jaccard\n" "similarity = Jaccard({k})"
            ).format(k=str(self.cfg.get("k", "")))
            if normalized:
                import_str = (
                    "\nfrom dtale.column_builders import SimilarityNormalizeWrapper\n"
                    "{base_import}\nsimilarity = SimilarityNormalizeWrapper(similarity)"
                ).format(base_import=base_import)
            else:
                import_str = "\n{}".format(base_import)

        return (
            "{import_str}\n"
            "distances = data[['{left}', '{right}']].fillna('').apply(lambda rec: similarity.distance(*rec))\n"
            "df.loc[:, '{name}'] = pd.Series(distances, index=data.index, name='{name}')"
        ).format(name=self.name, left=left_col, right=right_col, import_str=import_str)


class StandardizedColumnBuilder(object):
    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

    def build_column(self, data):
        col, algo = (self.cfg.get(p) for p in ["col", "algo"])
        if algo == "robust":
            try:
                from sklearn.preprocessing import RobustScaler

                transformer = RobustScaler()
            except ImportError:
                raise Exception(
                    "You must have at least scikit-learn 0.17.0 installed in order to use the RobustScaler!"
                )
        elif algo == "quantile":
            try:
                from sklearn.preprocessing import QuantileTransformer

                transformer = QuantileTransformer()
            except ImportError:
                raise Exception(
                    "You must have at least scikit-learn 0.19.0 installed in order to use the QuantileTransformer!"
                )
        elif algo == "power":
            try:
                from sklearn.preprocessing import PowerTransformer

                transformer = PowerTransformer(method="yeo-johnson", standardize=True)
            except ImportError:
                raise Exception(
                    "You must have at least scikit-learn 0.20.0 installed in order to use the PowerTransformer!"
                )
        standardized = transformer.fit_transform(data[[col]]).reshape(-1)
        return pd.Series(standardized, index=data.index, name=self.name)

    def build_code(self):
        col, algo = (self.cfg.get(p) for p in ["col", "algo"])
        import_str = ""
        if algo == "robust":
            import_str = (
                "\nfrom sklearn.preprocessing import RobustScaler\n"
                "transformer = RobustScaler()"
            )
        elif algo == "quantile":
            import_str = (
                "\nfrom sklearn.preprocessing import QuantileTransformer\n"
                "transformer = QuantileTransformer()"
            )
        elif algo == "power":
            import_str = (
                "\nfrom sklearn.preprocessing import PowerTransformer\n"
                "transformer = PowerTransformer(method='yeo-johnson', standardize=True)"
            )

        return (
            "{import_str}\n"
            "standardized = transformer.fit_transform(data[['{col}']]).reshape(-1)\n"
            "df.loc[:, '{name}'] = pd.Series(standardized, index=data.index, name='{name}')"
        ).format(name=self.name, col=col, import_str=import_str)


class EncoderColumnBuilder(object):
    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

    def build_column(self, data):
        col, algo = (self.cfg.get(p) for p in ["col", "algo"])
        if algo == "one_hot":
            return pd.get_dummies(data, columns=[col], drop_first=True)
        elif algo == "ordinal":
            try:
                from sklearn.preprocessing import OrdinalEncoder

                is_nan = data[col].isnull()
                ordinals = (
                    OrdinalEncoder()
                    .fit_transform(data[[col]].astype("str"))
                    .reshape(-1)
                )
                return pd.Series(ordinals, index=data.index, name=self.name).where(
                    ~is_nan, 0
                )
            except ImportError:
                raise Exception(
                    "You must have at least scikit-learn 0.20.0 installed in order to use the OrdinalEncoder!"
                )
        elif algo == "label":
            try:
                from sklearn.preprocessing import LabelEncoder

                is_nan = data[col].isnull()
                labels = LabelEncoder().fit_transform(data[col].astype("str"))
                return pd.Series(labels, index=data.index, name=self.name).where(
                    ~is_nan, 0
                )
            except ImportError:
                raise Exception(
                    "You must have at least scikit-learn 0.12.0 installed in order to use the LabelEncoder!"
                )
        elif algo == "feature_hasher":
            try:
                from sklearn.feature_extraction import FeatureHasher

                n = int(self.cfg.get("n"))
                features = (
                    FeatureHasher(n_features=n, input_type="string")
                    .transform(data[[col]].fillna("nan").astype("str").values)
                    .toarray()
                )
                features = pd.DataFrame(features, index=data.index)
                features.columns = [
                    "{}_{}".format(col, col2) for col2 in features.columns
                ]
                return features
            except ImportError:
                raise Exception(
                    "You must have at least scikit-learn 0.13.0 installed in order to use the FeatureHasher!"
                )
        raise NotImplementedError("{} not implemented yet!".format(algo))

    def build_code(self):
        col, algo = (self.cfg.get(p) for p in ["col", "algo"])
        if algo == "one_hot":
            return (
                "dummies = pd.get_dummies(df, columns=['{col}'], drop_first=True)\n"
                "for i in range(len(dummies.columns)):\n"
                "\tnew_col = dummies.iloc[:, i]\n"
                "\tdf.loc[:, str(new_col.name)] = new_col"
            ).format(col=col)
        elif algo == "ordinal":
            return (
                "\nfrom sklearn.preprocessing import OrdinalEncoder\n"
                "is_nan = df['{col}'].isnull()\n"
                "ordinals = OrdinalEncoder().fit_transform(df[['{col}']]).reshape(-1)\n"
                "df.loc[:, '{name}'] = pd.Series(ordinals, index=df.index, name='{name}').where(~is_nan, 0)"
            ).format(col=col, name=self.name)
        elif algo == "label":
            return (
                "\nfrom sklearn.preprocessing import LabelEncoder\n"
                "is_nan = df['{col}'].isnull()"
                "labels = OrdinalEncoder().fit_transform(df['{col}'])\n"
                "df.loc[:, '{name}'] = pd.Series(labels, index=df.index, name='{name}').where(~is_nan, 0)"
            ).format(col=col, name=self.name)
        elif algo == "feature_hasher":
            return (
                "\nfrom sklearn.feature_extraction import FeatureHasher\n"
                "hasher = FeatureHasher(n_features={n}, input_type='string')\n"
                "features = hasher.transform(data['{col}'].fillna('nan').astype('str')).toarray()\n"
                "features = pd.DataFrame(features, index=df.index)\n"
                "features.columns = ['{col}_' + col2 for col2 in features.columns]\n"
                "for i in range(len(features.columns)):\n"
                "\tnew_col = features.iloc[:, i]\n"
                "\tdf.loc[:, str(new_col.name)] = new_col"
            ).format(n=self.cfg.get("n"), col=col)
        return ""


printable = r"\w \!\"#\$%&\'\(\)\*\+,\-\./:;<»«؛،ـ\=>\?@\[\\\]\^_\`\{\|\}~"


def get_cleaner_configs():
    return [
        dict(
            value="drop_multispace", label=text("Replace Multi-Space w/ Single-Space")
        ),
        dict(value="drop_punctuation", label=text("Remove Punctuation")),
        dict(value="stopwords", label=text("Drop Stop Words")),
        dict(value="nltk_stopwords", label=text("Drop NLTK Stop Words")),
        dict(value="drop_numbers", label=text("Remove Numbers")),
        dict(value="keep_alpha", label=text("Keep Only Alpha")),
        dict(value="normalize_accents", label=text("Normalize Accent Characters")),
        dict(value="drop_all_space", label=text("Remove Spaces")),
        dict(value="drop_repeated_words", label=text("Drop Repeated Words")),
        dict(
            value="add_word_number_space",
            label=text("Add Space Between Word and Numbers"),
        ),
        dict(value="drop_repeated_chars", label=text("Remove Repeated Chars")),
        dict(value="update_case", label=text("Update Word Case")),
        dict(
            value="space_vals_to_empty",
            label=text("Update Space Values to Empty String"),
        ),
        dict(value="hidden_chars", label=text("Remove Hidden Characters")),
        dict(value="replace_hyphen_w_space", label=text("Replace Hyphens w/ Space")),
    ]


def clean(s, cleaner, cfg):
    replace_kwargs = {}
    if pandas_util.is_pandas2():
        replace_kwargs["regex"] = True

    if cleaner == "drop_multispace":
        return s.str.replace(r"[ ]+", " ", **replace_kwargs)
    elif cleaner == "drop_punctuation":
        if six.PY3:
            return apply(
                s, lambda x: x.translate(str.maketrans("", "", string.punctuation))
            )
        return apply(s, lambda x: x.translate(None, string.punctuation))
    elif cleaner == "stopwords":
        stopwords = cfg.get("stopwords") or []

        def clean_stopwords(x):
            return " ".join([w for w in x.split(" ") if w not in stopwords])

        return apply(s, clean_stopwords)
    elif cleaner == "nltk_stopwords":
        language = cfg.get("language") or "english"
        try:
            import nltk

            nltk.download("stopwords")
            nltk.download("punkt")

            nltk_stopwords_set = set(nltk.corpus.stopwords.words(language))

            def clean_nltk_stopwords(x):
                return " ".join(
                    [
                        w
                        for w in nltk.tokenize.word_tokenize(x)
                        if w not in nltk_stopwords_set
                    ]
                )

            return apply(s.fillna(""), clean_nltk_stopwords)
        except ImportError:
            raise Exception(
                "You must install the 'nltk' package in order to use this cleaner!"
            )
    elif cleaner == "drop_numbers":
        return s.str.replace(r"[\d]+", "", **replace_kwargs)
    elif cleaner == "keep_alpha":
        return apply(s, lambda x: "".join(c for c in x if c.isalpha()))
    elif cleaner == "normalize_accents":
        normalizer = "NFKD"
        if not six.PY3:
            normalizer = normalizer.decode("utf-8")
        return apply(
            s,
            lambda x: unicodedata.normalize(
                normalizer,
                "{}".format(x) if six.PY3 else "{}".format(x).decode("utf-8"),
            )
            .encode("ASCII", "ignore")
            .decode("utf-8"),
        )
    elif cleaner == "drop_all_space":
        return s.str.replace(r"[ ]+", "", **replace_kwargs)
    elif cleaner == "drop_repeated_words":

        def drop_repeats(val):
            def _load():
                val_segs = val.split(" ")
                for i, v2 in enumerate(val_segs):
                    if i == 0:
                        yield v2
                    elif val_segs[i - 1] != v2:
                        yield v2

            return " ".join(list(_load()))

        return apply(s, drop_repeats)
    elif cleaner == "add_word_number_space":
        return s.str.replace(r"(\d+(\.\d+)?)", r" \1 ", **replace_kwargs)
    elif cleaner == "drop_repeated_chars":

        def drop_repeats(val):
            def _load():
                for i, v2 in enumerate(val):
                    if i == 0:
                        yield v2
                    elif val[i - 1] != v2:
                        yield v2

            return "".join(list(_load()))

        return apply(s, drop_repeats)
    elif cleaner == "update_case":
        case = cfg.get("caseType")
        return getattr(s.str, case)()
    elif cleaner == "space_vals_to_empty":
        return s.str.replace(r"[ ]+", "", **replace_kwargs)
    elif cleaner == "underscore_to_space":
        return s.str.replace("_", " ", **replace_kwargs)
    elif cleaner == "hidden_chars":
        return s.str.replace(r"[^{}]+".format(printable), "")
    elif cleaner == "replace_hyphen_w_space":
        return s.str.replace(r"[‐᠆﹣－⁃−\-]+", " ", **replace_kwargs)
    return s


def clean_code(cleaner, cfg):
    replace_kwargs = ""
    if pandas_util.is_pandas2():
        replace_kwargs = ", regex = True"

    if cleaner == "drop_multispace":
        return ["s = s.str.replace(r'[ ]+', ' '{})".format(replace_kwargs)]
    elif cleaner == "drop_punctuation":
        if six.PY3:
            return [
                "s = s.apply(lambda x: x.translate(str.maketrans('', '', string.punctuation))"
            ]
        return ["s = s.apply(lambda x: x.translate(None, string.punctuation))"]
    elif cleaner == "stopwords":
        stopwords = cfg.get("stopwords") or []
        return [
            "def clean_stopwords(x):",
            "\treturn ' '.join([w for w in x.split(' ') if w not in ['{}'])".format(
                "','".join(stopwords)
            ),
            "s = s.apply(clean_stopwords)",
        ]
    elif cleaner == "nltk_stopwords":
        return [
            "import nltk\n",
            "nltk.download('stopwords')" "nltk.download('punkt')\n",
            "nltk_stopwords_set = set(nltk.corpus..words('{}'))\n".format(
                cfg.get("language") or "english"
            ),
            "def clean_nltk_stopwords(x):",
            "\treturn ' '.join(",
            "\t\t[w for w in nltk.tokenize.word_tokenize(x) if w not in nltk_stopwords_set]",
            "\t)",
            "s = s.apply(clean_nltk_stopwords)",
        ]
    elif cleaner == "drop_numbers":
        return ["""s = s.str.replace(r'[\\d]+', ''{})""".format(replace_kwargs)]
    elif cleaner == "keep_alpha":
        return ["s = s.apply(lambda x: ''.join(c for c in x if c.isalpha()))"]
    elif cleaner == "normalize_accents":
        return [
            "import unicodedata\n",
            "s = s.apply(",
            "\tlambda x: unicodedata.normalize('NFKD', u'{}'.format(x)).encode('ASCII', 'ignore').decode('utf-8')",
            ")",
        ]
    elif cleaner == "drop_all_space":
        return ["s.str.replace(r'[ ]+', ''{})".format(replace_kwargs)]
    elif cleaner == "drop_repeated_words":
        return [
            "def drop_repeated_words(val):",
            "\tdef _load():",
            "\t\tval_segs = val.split(' ')",
            "\t\t\tfor i, v2 in enumerate(val_segs):",
            "\t\t\t\tif i == 0:",
            "\t\t\t\t\tyield v2",
            "\t\t\t\telif val_segs[i - 1] != v2:",
            "\t\t\t\t\tyield v2",
            "\treturn ' '.join(list(_load()))",
            "s = s.apply(drop_repeated_words)",
        ]
    elif cleaner == "add_word_number_space":
        return [
            "s = s.str.replace(r'(\\d+(\\.\\d+)?)', r' \\1 '{})".format(replace_kwargs)
        ]
    elif cleaner == "drop_repeated_chars":
        return [
            "def drop_repeated_chars(val):",
            "\tdef _load():",
            "\t\tfor i, v2 in enumerate(val):",
            "\t\t\tif i == 0:",
            "\t\t\t\tyield v2",
            "\t\t\telif val[i - 1] != v2:",
            "\t\t\t\tyield v2",
            "\treturn ''.join(list(_load()))",
            "s = s.apply(drop_repeated_chars)",
        ]
    elif cleaner == "update_case":
        case = cfg.get("caseType")
        return ["s = s.str.{}()".format(case)]
    elif cleaner == "space_vals_to_empty":
        return ["s = s.str.replace(r'[ ]+', ''{})".format(replace_kwargs)]
    elif cleaner == "underscore_to_space":
        return ["s = s.str.replace('_', ' '{})".format(replace_kwargs)]
    elif cleaner == "hidden_chars":
        return [
            "printable = r'\\w \\!\"#\\$%&'\\(\\)\\*\\+,\\-\\./:;<»«؛،ـ\\=>\\?@\\[\\\\\\]\\^_\\`\\{\\|\\}~'",
            "s = s.str.replacer(r'[^{}]+'.format(printable), '')",
        ]
    elif cleaner == "replace_hyphen_w_space":
        return "s = s.str.replace(r'[‐᠆﹣－⁃−-]+', ' '{})".format(replace_kwargs)
    return []


class CleaningColumnBuilder(object):
    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

    def build_column(self, data):
        col, cleaners = (self.cfg.get(p) for p in ["col", "cleaners"])
        s = data[col]
        for cleaner in cleaners:
            s = clean(s, cleaner, self.cfg)
        return s

    def build_code(self):
        col, cleaners = (self.cfg.get(p) for p in ["col", "cleaners"])
        code = ["s= ddf['{col}']".format(col=col)]
        for cleaner in cleaners:
            code += clean_code(cleaner, self.cfg)
        code.append(
            "df.loc[:, '{name}'] = pd.Series(s, index=df.index, name='{name}')".format(
                name=self.name
            )
        )
        return code


class DiffColumnBuilder(object):
    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

    def build_column(self, data):
        col, periods = (self.cfg.get(p) for p in ["col", "periods"])
        return pd.Series(data[col].diff(int(periods)), index=data.index, name=self.name)

    def build_code(self):
        col, periods = (self.cfg.get(p) for p in ["col", "periods"])
        return "df.loc[:, '{name}'] = pd.Series(df['{col}'].diff({periods}), index=df.index, name='{name}')".format(
            name=self.name, col=col, periods=periods
        )


class TimeseriesDataSlopeBuilder(object):
    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

    def build_column(self, data):
        col = self.cfg.get("col")
        diffs = data[col].diff().bfill()
        diffs.loc[diffs < 0] = -1
        g = (~(diffs == diffs.shift(1))).cumsum()
        return pd.Series(g, index=data.index, name=self.name)

    def build_code(self):
        col = self.cfg.get("col")
        return [
            "diffs = df['{}'].diff().bfill()".format(col),
            "diffs.loc[diffs < 0] = -1",
            "g = (~(diffs == diffs.shift(1))).cumsum()",
            "df.loc[:, '{name}'] = pd.Series(g, index=df.index, name='{name}')".format(
                name=self.name
            ),
        ]


class RollingBuilder(object):
    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

    def _get_props(self):
        col, comp, window, center = (
            self.cfg.get(p) for p in ["col", "comp", "window", "center"]
        )
        rolling_kwargs = {
            p: self.cfg[p]
            for p in ["min_periods", "center", "win_type", "on", "closed"]
            if self.cfg.get(p) is not None
        }
        window = int(window)
        if rolling_kwargs.get("min_periods"):
            rolling_kwargs["min_periods"] = int(rolling_kwargs["min_periods"])
        return col, comp, window, center, rolling_kwargs

    def build_column(self, data):
        col, comp, window, center, rolling_kwargs = self._get_props()
        if rolling_kwargs.get("on"):
            vals = data[[col, rolling_kwargs["on"]]].rolling(window, **rolling_kwargs)[
                col
            ]
        else:
            vals = data[col].rolling(window, **rolling_kwargs)
        vals = getattr(vals, comp)()
        return pd.Series(vals, index=data.index, name=self.name)

    def build_code(self):
        col, comp, window, center, rolling_kwargs = self._get_props()
        rolling_kwargs_str = []
        if rolling_kwargs.get("min_periods"):
            rolling_kwargs_str.append(
                "min_periods={}".format(rolling_kwargs["min_periods"])
            )
        if rolling_kwargs.get("center"):
            rolling_kwargs_str.append("center=True")
        for p in ["win_type", "on", "closed"]:
            if rolling_kwargs.get(p):
                rolling_kwargs_str.append("{}='{}'".format(p, rolling_kwargs[p]))
        rolling_kwargs_str = ", ".join(rolling_kwargs_str)
        if rolling_kwargs.get("on"):
            code = [
                "rolling_vals = df[['{}', '{}']].rolling(",
                "\t{}{}".format(
                    window,
                    "" if not rolling_kwargs_str else ", {}".format(rolling_kwargs_str),
                ),
                ").{}()['{}']".format(comp, col),
            ]
        else:
            code = [
                "rolling_vals = df['{col}'].rolling({window}{kwargs}).{comp}()".format(
                    col=col,
                    window=window,
                    kwargs=(
                        ""
                        if not rolling_kwargs_str
                        else ", {}".format(rolling_kwargs_str)
                    ),
                    comp=comp,
                )
            ]
        return code + [
            "df.loc[:, '{name}'] = pd.Series(colling_vals, index=df.index, name='{name}')".format(
                name=self.name
            )
        ]


class ExponentialSmoothingBuilder(object):
    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

    def build_column(self, data):
        col, alpha = (self.cfg.get(p) for p in ["col", "alpha"])
        alpha = float(alpha)
        s = data[col].values
        result = [s[0]]
        for n in range(1, len(s)):
            result.append(alpha * s[n] + (1 - alpha) * result[n - 1])
        return pd.Series(result, index=data.index, name=self.name)

    def build_code(self):
        col, alpha = (self.cfg.get(p) for p in ["col", "alpha"])
        return [
            "s = df['{}'].values".format(col),
            "result = [s[0]]",
            "for n in range(1, len(s)):",
            "\tresult.append({alpha} * s[n] + (1 - {alpha}) * result[n - 1])".format(
                alpha=alpha
            ),
            "df.loc[:, '{name}'] = pd.Series(result, index=df.index, name='{name}')".format(
                name=self.name
            ),
        ]


class CumsumColumnBuilder(object):
    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

    def build_column(self, data):
        cols = self.cfg.get("cols")
        group = self.cfg.get("group")
        if group:
            s = data.groupby(group)[cols]
        else:
            s = data[cols]
        kwargs = {}
        if not pandas_util.is_pandas3():
            kwargs["axis"] = 0
        s = s.cumsum(**kwargs)
        s.columns = ["{}{}".format(col, self.name) for col in s.columns]
        return pd.DataFrame(s, index=data.index)

    def build_code(self):
        cols = self.cfg.get("cols")
        group = self.cfg.get("group")
        group_code = ""
        if group:
            group_code = ".groupby(['{}'])".format("','".join(group))
        return (
            "cumsum = pd.DataFrame(\n"
            "\t(data{group}['{cols}'].cumsum(axis=0), index=data.index\n"
            ")\n"
            "for i in range(len(cumsum.columns)):\n"
            "\tnew_col = cumsum.iloc[:, i]\n"
            "\tdf.loc[:, str(new_col.name) + '{name}'] = new_col"
        ).format(name=self.name, cols="','".join(cols), group=group_code)


class ShiftBuilder(object):
    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

    def build_column(self, data):
        col, periods, fill_value, dtype = (
            self.cfg.get(p) for p in ["col", "periods", "fillValue", "dtype"]
        )
        kwargs = {}
        if fill_value is not None and pandas_util.check_pandas_version("0.24.0"):
            fill_formatter = find_dtype_formatter(dtype)
            kwargs["fill_value"] = fill_formatter(fill_value)
        return pd.Series(
            data[col].shift(periods or 1, **kwargs), index=data.index, name=self.name
        )

    def build_code(self):
        col, periods, fill_value, dtype = (
            self.cfg.get(p) for p in ["col", "periods", "fillValue", "dtype"]
        )
        kwargs = ""
        if fill_value is not None and pandas_util.check_pandas_version("0.24.0"):
            if classify_type(dtype) == "S":
                kwargs = ", fill_value='{}'".format(fill_value)
            else:
                kwargs = ", fill_value={}".format(fill_value)
        return (
            "df.loc[:, '{name}'] = pd.Series(\n"
            "\t(data['{col}'].shift({periods}{kwargs}), index=data.index, name='{name}'\n"
            ")"
        ).format(name=self.name, col=col, periods=periods, kwargs=kwargs)


class ExpandingBuilder(object):
    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

    def build_column(self, data):
        col, periods, agg = (self.cfg.get(p) for p in ["col", "periods", "agg"])
        return pd.Series(
            getattr(data[col].expanding(periods or 1), agg)(),
            index=data.index,
            name=self.name,
        )

    def build_code(self):
        col, periods, agg = (self.cfg.get(p) for p in ["col", "periods", "agg"])
        return (
            "df.loc[:, '{name}'] = pd.Series(\n"
            "\t(data['{col}'].expanding({periods}).{agg}(), index=data.index, name='{name}'\n"
            ")"
        ).format(name=self.name, col=col, periods=periods or 1, agg=agg)


class SubstringBuilder(object):
    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

    def build_column(self, data):
        col, start, end = (self.cfg.get(p) for p in ["col", "start", "end"])
        start = int(start)
        end = int(end)
        s = data[col].str[:end] if start == 0 else data[col].str.slice(start, end)
        return pd.Series(s, index=data.index, name=self.name)

    def build_code(self):
        col, start, end = (self.cfg.get(p) for p in ["col", "start", "end"])
        if start == 0:
            return (
                "df.loc[:, '{name}'] = pd.Series(data['{col}'].str[:{end}], index=data.index, name='{name}')"
            ).format(name=self.name, col=col, end=end)
        return (
            "df.loc[:, '{name}'] = pd.Series(\n"
            "\t(data['{col}'].str.slice({start}, {end}), index=data.index, name='{name}'\n"
            ")"
        ).format(name=self.name, col=col, start=start, end=end)


class SplitBuilder(object):
    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

    def build_column(self, data):
        col, delimiter = (self.cfg.get(p) for p in ["col", "delimiter"])
        splits = data[col].str.split(delimiter, expand=True)
        splits.columns = ["{}_{}".format(self.name, col2) for col2 in splits.columns]
        return splits

    def build_code(self):
        col, delimiter = (self.cfg.get(p) for p in ["col", "delimiter"])
        return (
            "splits = data['{col}'].str.split('{delimiter}'), expand=True)\n"
            "splits.columns = ['{name}_' + col2 for col2 in splits.columns]\n"
            "for i in range(len(splits.columns)):\n"
            "\tnew_col = splits.iloc[:, i]\n"
            "\tdf.loc[:, str(new_col.name)] = new_col"
        ).format(col=col, delimiter=delimiter, name=self.name)



import pandas as pd
import sqlite3

from dtale.app import show
from dtale.cli.clickutils import get_loader_options

"""
  IMPORTANT!!! These global variables are required for building any customized CLI loader.
  When build_loaders runs startup it will search for any modules containing the global variable LOADER_KEY.
"""
LOADER_KEY = "sqlite"
LOADER_PROPS = [
    dict(name="path", help="path to sqlite file"),
    dict(name="table", help="the table to load to D-Tale --sqlite-path"),
]


# IMPORTANT!!! This function is required if you would like to be able to use this loader from the back-end.
def show_loader(**kwargs):
    return show(data_loader=lambda: loader_func(**kwargs), **kwargs)


def loader_func(**kwargs):
    if "table" not in kwargs:
        raise Exception("You must specify a table name in order to use sqlite loader!")
    path = kwargs.pop("path")
    cnx = sqlite3.connect(path)
    return pd.read_sql_query("SELECT * FROM {}".format(kwargs.pop("table")), cnx)


# IMPORTANT!!! This function is required for building any customized CLI loader.
def find_loader(kwargs):
    """
    SQLite implementation of data loader which will return a function if any of the
    `click` options based on LOADER_KEY & LOADER_PROPS have been used, otherwise return None

    :param kwargs: Optional keyword arguments to be passed from `click`
    :return: data loader function for SQLite implementation
    """
    sqlite_opts = get_loader_options(LOADER_KEY, LOADER_PROPS, kwargs)
    if len([f for f in sqlite_opts.values() if f]):

        def _sqlite_loader():
            return loader_func(**sqlite_opts)

        return _sqlite_loader
    return None



import json

import dash_bootstrap_components as dbc
import dash_daq as daq
import os
import plotly
from six import PY3

from dtale.dash_application import dcc, html
import dtale.dash_application.components as dash_components
import dtale.dash_application.custom_geojson as custom_geojson
import dtale.dash_application.extended_aggregations as extended_aggregations
import dtale.global_state as global_state
from dtale.charts.utils import (
    AGGS,
    ANIMATION_CHARTS,
    ANIMATE_BY_CHARTS,
    YAXIS_CHARTS,
    ZAXIS_CHARTS,
    NON_EXT_AGGREGATION,
    build_final_cols,
    find_group_vals,
)
from dtale.column_builders import get_cleaner_configs
from dtale.constants import CHART_JOINER_CHAR
from dtale.dash_application.layout.utils import (
    build_input,
    build_option,
    build_tab,
    build_cols,
    build_hoverable,
    build_selections,
    show_style,
    FREQS,
    FREQ_LABELS,
)
from dtale.translations import text
from dtale.query import build_query, handle_predefined, inner_build_query, run_query
from dtale.utils import (
    ChartBuildingError,
    classify_type,
    coord_type,
    get_dtypes,
    is_app_root_defined,
    make_list,
    parse_version,
)


def test_plotly_version(version_num):
    if "unknown" in plotly.__version__:
        return True
    return parse_version(plotly.__version__) >= parse_version(version_num)


def get_dashbio():
    import dash_bio as dashbio  # noqa: F401

    return dashbio


def has_dashbio():
    try:
        get_dashbio()
        return True
    except ImportError:
        return False


def base_layout(app_root, **kwargs):
    """
    Base layout to be returned by :meth:`dtale.dash_application.views.DtaleDash.interpolate_index`

    :param kwargs: Optional keyword arguments to be passed to 'dash.Dash.interplolate_index'
    :type kwargs: dict
    :return: HTML
    :rtype: str
    """
    webroot_html = ""
    favicon_path = "../../dtale/static/images/favicon.png"
    if is_app_root_defined(app_root):
        webroot_html = """
        <script type="text/javascript">
            window.resourceBaseUrl = '{app_root}';
        </script>
        """.format(app_root=app_root)
        favicon_path = "{}/dtale/static/images/favicon.png".format(app_root)
    grid_links = []
    for id in global_state.keys():
        label = global_state.get_name(id)
        if label:
            label = " ({})".format(label)
        else:
            label = ""
        grid_links.append(
            (
                """<a href="../main/{id}" class="dropdown-item data-grid-link" target="{target}">"""
                """<span class="ml-3">{id}{label}</span>"""
                """</a>"""
            ).format(
                id=id,
                label=label,
                target=(
                    "_self" if os.environ.get("VSCODE_PID") is not None else "_blank"
                ),
            )
        )
    language = global_state.get_app_settings()["language"]
    language_links = []
    for value, label in [("en", "English"), ("cn", "Chinese"), ("pt", "Portuguese")]:
        if value == language:
            language_links.append(
                """<span class="dropdown-item active">{}</span>""".format(
                    "&#10003; {}".format(text(label))
                )
            )
        else:
            language_links.append(
                """<a href="{}" class="dropdown-item lang-link">{}</a>""".format(
                    value, text(label)
                )
            )

    main_title = global_state.get_app_settings().get("main_title") or "D-TALE"
    main_title_font = global_state.get_app_settings().get("main_title_font")
    main_title_class = "title-font-base"
    main_title_style = ""
    if main_title_font:
        main_title_style = "font-family: {main_title_font}".format(
            main_title_font=main_title_font
        )
    else:
        main_title_class = "title-font title-font-base"
    page_title = global_state.get_app_settings().get("main_title") or "D-Tale"

    return """
        <!DOCTYPE html>
        <html>
            <head>
                {webroot_html}
                {metas}
                <title>{page_title} Charts</title>
                <link rel="shortcut icon" href="{favicon_path}">
                {css}
            </head>
            <body class="{theme}-mode">
                <header class="app-header">
                    <span class="{main_title_class}" style="{main_title_style}">{main_title}</span>
                    <span style="font-size: 16px" class="pl-5 mt-4">{title}</span>
                    <nav class="app-header__nav--secondary">
                        <ul class="nav-menus">
                            <li>
                                <div class="dropdown">
                                    <span class="dropbtn nav-link dropdown-toggle">
                                        {back_to_data}
                                    </span>
                                    <div class="dropdown-content">{grid_links}</div>
                                </div>
                            </li>
                            <li>
                                <div class="dropdown">
                                    <span class="dropbtn nav-link dropdown-toggle">
                                        {languages}
                                    </span>
                                    <div class="dropdown-content">{language_links}</div>
                                </div>
                            </li>
                        </ul>
                    </nav>
                </header>
                <div class="container-fluid charts">
                    {app_entry}
                </div>
                <footer>
                    {config}
                    {scripts}
                    <script type="text/javascript">
                        const pathSegs = window.location.pathname.split('/');
                        const dataId = pathSegs[pathSegs.length - 1];
                        const backToData = () => window.open('{app_root}/dtale/main/' + dataId);
                    </script>
                    {renderer}
                    {css}
                </footer>
            </body>
        </html>
    """.format(
        metas=kwargs["metas"],
        css=kwargs["css"],
        app_entry=kwargs["app_entry"],
        config=kwargs["config"],
        scripts=kwargs["scripts"],
        renderer=kwargs["renderer"],
        webroot_html=webroot_html,
        app_root=app_root if is_app_root_defined(app_root) else "",
        favicon_path=favicon_path,
        title=text("Charts"),
        back_to_data=text("Back To Data"),
        grid_links="".join(grid_links),
        languages=text("Languages"),
        language_links="".join(language_links),
        main_title=main_title,
        main_title_class=main_title_class,
        main_title_style=main_title_style,
        page_title=page_title,
        theme=global_state.get_app_settings()["theme"],
    )


CHARTS = [
    dict(value="line"),
    dict(value="bar"),
    dict(value="scatter"),
    dict(value="pie"),
    dict(value="wordcloud"),
    dict(value="heatmap"),
    dict(value="3d_scatter", label="3D Scatter"),
    dict(value="surface"),
    dict(value="maps"),
    dict(value="candlestick"),
    dict(value="treemap"),
    dict(value="funnel"),
]
if has_dashbio():
    CHARTS.append(dict(value="clustergram"))
CHARTS.append(dict(value="pareto"))
CHARTS.append(dict(value="histogram"))

CHART_INPUT_SETTINGS = {
    "line": dict(
        x=dict(type="single"),
        y=dict(type="multi"),
        z=dict(display=False),
        group=dict(display=True, type="single"),
    ),
    "bar": dict(
        x=dict(type="single"),
        y=dict(type="multi"),
        z=dict(display=False),
        group=dict(display=True, type="single"),
    ),
    "scatter": dict(
        x=dict(type="single"),
        y=dict(type="multi"),
        z=dict(display=False),
        group=dict(display=True, type="single"),
    ),
    "pie": dict(
        x=dict(type="single"),
        y=dict(type="multi"),
        z=dict(display=False),
        group=dict(display=True, type="single"),
    ),
    "wordcloud": dict(
        x=dict(type="single"),
        y=dict(type="multi"),
        z=dict(display=False),
        group=dict(display=True, type="single"),
    ),
    "heatmap": dict(
        x=dict(type="single"),
        y=dict(type="single"),
        z=dict(display=True, type="single"),
        group=dict(display=False),
    ),
    "3d_scatter": dict(
        x=dict(type="single"),
        y=dict(type="single"),
        z=dict(display=True, type="single"),
        group=dict(display=True),
    ),
    "surface": dict(
        x=dict(type="single"),
        y=dict(type="single"),
        z=dict(display=True, type="single"),
        group=dict(display=False),
    ),
    "maps": dict(
        x=dict(display=False),
        y=dict(display=False),
        z=dict(display=False),
        group=dict(display=False),
        map_group=dict(display=True),
    ),
    "candlestick": dict(
        x=dict(display=False),
        y=dict(display=False),
        z=dict(display=False),
        group=dict(display=False),
        map_group=dict(display=False),
        cs_group=dict(display=True),
    ),
    "treemap": dict(
        x=dict(display=False),
        y=dict(display=False),
        z=dict(display=False),
        group=dict(display=False),
        map_group=dict(display=False),
        cs_group=dict(display=False),
        treemap_group=dict(display=True),
    ),
    "funnel": dict(
        x=dict(display=False),
        y=dict(display=False),
        z=dict(display=False),
        group=dict(display=False),
        map_group=dict(display=False),
        cs_group=dict(display=False),
        treemap_group=dict(display=False),
        funnel_group=dict(display=True),
    ),
    "clustermap": dict(
        x=dict(display=False),
        y=dict(display=False),
        z=dict(display=False),
        group=dict(display=False),
        map_group=dict(display=False),
        cs_group=dict(display=False),
        treemap_group=dict(display=False),
        funnel_group=dict(display=False),
        clustergram_group=dict(display=True),
    ),
    "pareto": dict(
        x=dict(display=False),
        y=dict(display=False),
        z=dict(display=False),
        group=dict(display=False),
        map_group=dict(display=False),
        cs_group=dict(display=False),
        treemap_group=dict(display=False),
        funnel_group=dict(display=False),
        clustergram_group=dict(display=False),
    ),
    "histogram": dict(
        x=dict(display=False),
        y=dict(display=False),
        z=dict(display=False),
        group=dict(display=False),
        map_group=dict(display=False),
        cs_group=dict(display=False),
        treemap_group=dict(display=False),
        funnel_group=dict(display=False),
        clustergram_group=dict(display=False),
        histogram_group=dict(display=True),
    ),
}
LOAD_TYPES = ["random", "head", "tail", "stratified"]
MAP_TYPES = [
    dict(value="choropleth", image=True),
    dict(value="scattergeo", label="ScatterGeo", image=True),
    dict(value="mapbox", label="Detailed", image=True),
]
SCOPES = ["world", "usa", "europe", "asia", "africa", "north america", "south america"]
PROJECTIONS = [
    "equirectangular",
    "mercator",
    "orthographic",
    "natural earth",
    "kavrayskiy7",
    "miller",
    "robinson",
    "eckert4",
    "azimuthal equal area",
    "azimuthal equidistant",
    "conic equal area",
    "conic conformal",
    "conic equidistant",
    "gnomonic",
    "stereographic",
    "mollweide",
    "hammer",
    "transverse mercator",
    "albers usa",
    "winkel tripel",
    "aitoff",
    "sinusoidal",
]


def auto_load_msg():
    return text("auto_load_msg")


def load_type_msg():
    return [
        html.Div(
            [
                html.H3(
                    text("Load Types"), style=dict(display="inline"), className="pr-3"
                )
            ]
        ),
        html.Ul(
            [
                html.Li(
                    [
                        html.B(text("Random")),
                        html.Span(": {}".format(text("load_random"))),
                    ],
                    className="mb-0",
                ),
                html.Li(
                    [html.B(text("Head")), html.Span(": {}".format(text("load_head")))],
                    className="mb-0",
                ),
                html.Li(
                    [html.B(text("Tail")), html.Span(": {}".format(text("load_tail")))],
                    className="mb-0",
                ),
                html.Li(
                    [
                        html.B(text("Stratified")),
                        html.Span(": {}".format(text("load_stratified"))),
                    ],
                    className="mb-0",
                ),
            ],
            className="mb-0",
        ),
    ]


def bin_type_msg():
    return [
        html.Div(
            [
                html.H3(
                    text("Bin Types"), style=dict(display="inline"), className="pr-3"
                ),
                html.A(
                    "({})".format(text("Binning in Data Mining")),
                    href="https://www.geeksforgeeks.org/binning-in-data-mining/",
                ),
                html.Br(),
            ]
        ),
        html.Ul(
            [
                html.Li(
                    [
                        html.B(text("Equal Width")),
                        html.Span(
                            ": {}".format(text("the bins are of equal sized ranges"))
                        ),
                    ],
                    className="mb-0",
                ),
                html.Li(
                    [
                        html.B(text("Equal Frequency")),
                        html.Span(
                            ": {}".format(
                                text("the bins contain an equal amount of values")
                            )
                        ),
                    ],
                    className="mb-0",
                ),
            ],
            className="mb-0",
        ),
    ]


def build_img_src(proj, img_type="projections"):
    return "../static/images/{}/{}.png".format(img_type, "_".join(proj.split(" ")))


def build_proj_hover_children(proj):
    if proj is None:
        return None
    return [
        html.I(className="ico-help-outline", style=dict(color="white")),
        html.Div(
            [html.Div(proj), html.Img(src=build_img_src(proj))],
            className="hoverable__content",
            style=dict(width="auto"),
        ),
    ]


def build_proj_hover(proj):
    return html.Span(
        [
            text("Projection"),
            html.Div(
                build_proj_hover_children(proj),
                className="ml-3 hoverable",
                style=(
                    dict(display="none") if proj is None else dict(borderBottom="none")
                ),
                id="proj-hover",
            ),
        ],
        className="input-group-addon",
    )


def build_mapbox_token_children():
    from dtale.charts.utils import get_mapbox_token

    msg = text("To access additional styles enter a token here...")
    if get_mapbox_token() is None:
        msg = text("Change your token here...")
    return [
        html.I(className="ico-help-outline", style=dict(color="white")),
        html.Div(
            [
                html.Span("{}:".format("Mapbox Access Token")),
                dcc.Input(
                    id="mapbox-token-input",
                    type="text",
                    placeholder=msg,
                    className="form-control",
                    value="",
                    style={"lineHeight": "inherit"},
                ),
            ],
            className="hoverable__content",
            style=dict(width="20em", right="-1.45em"),
        ),
    ]


def build_mapbox_token_hover():
    return html.Span(
        [
            text("Style"),
            html.Div(
                build_mapbox_token_children(),
                className="ml-3 hoverable",
                style=dict(borderBottom="none"),
                id="token-hover",
            ),
        ],
        className="input-group-addon",
    )


def loc_mode_info():
    return {
        "ISO-3": dict(
            url=html.A(
                text("ISO-3"),
                href="https://en.wikipedia.org/wiki/ISO_3166-1_alpha-3",
                target="_blank",
            ),
            examples=["USA (United States)", "CAN (Canada)", "GBR (United Kingdom)"],
        ),
        "USA-states": dict(
            url=html.A(
                text("USA-states (use ISO)"),
                href="https://en.wikipedia.org/wiki/List_of_U.S._state_abbreviations",
                target="_blank",
            ),
            examples=["NY (New York)", "CA (California)", "MA (Massachusetts)"],
        ),
        "country names": dict(
            url=html.A(
                text("country names (case-insensitive)"),
                href="https://simple.wikipedia.org/wiki/List_of_countries",
                target="_blank",
            ),
            examples=["United States", "United Kingdom", "Germany"],
        ),
        "geojson-id": dict(
            label="Custom GeoJSON", desc=text("Load custom geojson into D-Tale")
        ),
    }


def build_loc_mode_hover_children(loc_mode):
    if loc_mode is None:
        return None
    loc_modes = loc_mode_info()
    loc_mode_cfg = loc_modes[loc_mode]
    url, examples, desc = (loc_mode_cfg.get(p) for p in ["url", "examples", "desc"])
    body = []
    if url is not None:
        body.append(
            html.Div([html.Span(text("View"), className="mr-3"), loc_mode_cfg["url"]])
        )
    if examples is not None:
        body.append(
            html.Ul(
                [
                    html.Li(e, style={"listStyleType": "disc"}, className="mb-3")
                    for e in loc_mode_cfg["examples"]
                ],
                className="pt-3 mb-0",
            )
        )
    if desc is not None:
        body.append(html.Span(desc))
    return [
        html.I(className="ico-help-outline", style=dict(color="white")),
        html.Div(
            body,
            className="hoverable__content build-code",
            style=dict(width="auto", whiteSpace="nowrap", left="-2em", top="auto"),
        ),
    ]


def build_loc_mode_hover(loc_mode):
    return html.Span(
        [
            html.Span(text("Location Mode"), style=dict(whiteSpace="pre-line")),
            html.Div(
                build_loc_mode_hover_children(loc_mode),
                className="ml-3 hoverable",
                style=(
                    dict(display="none")
                    if loc_mode is None
                    else dict(borderBottom="none")
                ),
                id="loc-mode-hover",
            ),
        ],
        className="input-group-addon pt-1 pb-0",
    )


JET = ["#000083", "#003CAA", "#05FFFF", "#FFFF00", "#FA0000", "#800000"]
REDS = ["#fff5f0", "#fdcab4", "#fc8a6a", "#f24632", "#bc141a", "#67000d"]
YLORRD = [
    "#ffffcc",
    "#ffeda0",
    "#fed976",
    "#feb24c",
    "#fd8d3c",
    "#fc4e2a",
    "#e31a1c",
    "#bd0026",
    "#800026",
]
YLGRBL = ["#ffffd9", "#d5efb3", "#73c9bc", "#1b99c2", "#1c4ea2", "#081d58"]
DEFAULT_CSALES = {"heatmap": JET, "maps": REDS, "3d_Scatter": YLORRD, "surface": YLGRBL}


def show_input_handler(chart_type):
    settings = CHART_INPUT_SETTINGS.get(chart_type or "line") or {}

    def _show_input(input_id, input_type="single"):
        cfg = settings.get(input_id, {})
        return cfg.get("display", True) and cfg.get("type", "single") == input_type

    return _show_input


def get_group_types(inputs, group_cols=None):
    def _flags(group_prop):
        final_group_cols = group_cols or make_list(inputs.get(group_prop))
        if len(final_group_cols):
            dtypes = global_state.get_dtypes(inputs["data_id"])
            for dtype_info in dtypes:
                if dtype_info["name"] in final_group_cols:
                    classifier = classify_type(dtype_info["dtype"])
                    if classifier == "F":
                        return ["bins"]
                    if classifier == "I":
                        return ["groups", "bins"]
                    return ["groups"]
            for fgc in final_group_cols:
                col, freq = fgc.split(CHART_JOINER_CHAR)
                col_exists = (
                    next(
                        (
                            dtype_info
                            for dtype_info in dtypes
                            if dtype_info["name"] == col
                        ),
                        None,
                    )
                    is not None
                )
                if col_exists and freq in FREQS:
                    return ["groups"]
        return []

    chart_type = inputs.get("chart_type")

    if show_input_handler(chart_type)("group"):
        return _flags("group")
    elif show_input_handler(chart_type)("map_group"):
        return _flags("map_group")
    elif show_input_handler(chart_type)("cs_group"):
        return _flags("cs_group")
    elif show_input_handler(chart_type)("treemap_group"):
        return _flags("treemap_group")
    elif show_input_handler(chart_type)("funnel_group"):
        return _flags("funnel_group")
    elif show_input_handler(chart_type)("clustergram_group"):
        return _flags("clustergram_group")
    elif show_input_handler(chart_type)("pareto_group"):
        return _flags("pareto_group")
    return False, False


def update_label_for_freq_and_agg(val):
    """
    Formats sub-values contained within 'val' to display date frequencies & aggregatioms if included.
        - (val=['a', 'b', 'c']) => 'a, b, c'
        - (val=['a||H', 'b||mean', 'c']) => 'a (Hour), Mean of b, c'
    """

    def _freq_handler(sub_val):
        selected_agg = None
        for agg in AGGS:
            if sub_val.endswith("{}{}".format(CHART_JOINER_CHAR, agg)):
                selected_agg = "{} of".format(text(AGGS[agg]))
                sub_val = sub_val.split("{}{}".format(CHART_JOINER_CHAR, agg))[0]
                break

        selected_freq = None
        for freq in FREQS:
            if sub_val.endswith("{}{}".format(CHART_JOINER_CHAR, freq)):
                col, freq = sub_val.split(CHART_JOINER_CHAR)
                if freq in FREQS:
                    sub_val = sub_val.split("{}{}".format(CHART_JOINER_CHAR, freq))[0]
                    if freq in FREQ_LABELS:
                        selected_freq = "({})".format(text(FREQ_LABELS[freq]))
                    break

        return " ".join(list(filter(None, [selected_agg, sub_val, selected_freq])))

    return ", ".join([_freq_handler(sub_val) for sub_val in make_list(val)])


def build_error(error, tb):
    """
    Returns error/traceback information in standard component with styling

    :param error: execption message
    :type error: str
    :param tb: tracebackF
    :type tb: str
    :return: error component
    :rtype: :dash:`dash_html_components.Div <dash-html-components/div>`
    """
    if isinstance(error, ChartBuildingError):
        if error.details:
            tb = error.details
        error = error.error
    return html.Div(
        [
            html.I(className="ico-error"),
            html.Span(str(error)),
            html.Div(html.Pre(str(tb)), className="traceback"),
        ],
        className="dtale-alert alert alert-danger",
    )


def build_input_options(df, extended_aggregation=[], **inputs):
    """
    Builds dropdown options for (X, Y, Z, Group, Barsort & Y-Axis Ranges) with filtering based on currently selected
    values for the following inputs: x, y, z, group.
    """
    chart_type, x, y, z, group = (
        inputs.get(p) for p in ["chart_type", "x", "y", "z", "group"]
    )
    col_opts = list(build_cols(df.columns, get_dtypes(df)))
    group_val, z_val = (None, z) if chart_type in ZAXIS_CHARTS else (group, None)
    x_options = [
        build_option(c, l)
        for c, l in col_opts
        if c not in build_selections(y, z_val, group_val)
    ]
    y_filter = build_selections(x, group_val, z_val)
    y_multi_options = [build_option(c, l) for c, l in col_opts if c not in y_filter]
    y_single_options = [build_option(c, l) for c, l in col_opts if c not in y_filter]
    z_options = [
        build_option(c)
        for c in df.columns
        if c not in build_selections(x, y, group_val)
    ]
    group_options = [
        build_option(c, l)
        for c, l in col_opts
        if c not in build_selections(x, y, z_val)
    ]
    final_cols = build_final_cols(
        y,
        z,
        inputs.get("agg"),
        extended_aggregation if chart_type not in NON_EXT_AGGREGATION else [],
    )
    barsort_options = [build_option(o) for o in build_selections(x, final_cols)]
    yaxis_options = [build_option(y2) for y2 in final_cols or []]
    return (
        x_options,
        y_multi_options,
        y_single_options,
        z_options,
        group_options,
        barsort_options,
        yaxis_options,
        [build_option(c, l) for c, l in col_opts],
    )


def build_map_options(
    df, type="choropleth", loc=None, lat=None, lon=None, map_val=None
):
    dtypes = get_dtypes(df)
    cols = sorted(dtypes.keys())
    lat_cols, lon_cols, str_cols, num_cols = [], [], [], []
    for c in cols:
        dtype = dtypes[c]
        classification = classify_type(dtype)
        if classification == "S":
            str_cols.append(c)
            continue
        if classification in ["F", "I"]:
            num_cols.append(c)
            coord = coord_type(df[c])
            if coord == "lat":
                lat_cols.append(c)
            elif coord == "lon":
                lon_cols.append(c)

    lat_options = [
        build_option(c) for c in lat_cols if c not in build_selections(lon, map_val)
    ]
    lon_options = [
        build_option(c) for c in lon_cols if c not in build_selections(lat, map_val)
    ]
    loc_options = [
        build_option(c) for c in str_cols if c not in build_selections(map_val)
    ]

    if type == "choropleth":
        val_options = [
            build_option(c) for c in num_cols if c not in build_selections(loc)
        ]
    else:
        val_options = [
            build_option(c) for c in num_cols if c not in build_selections(lon, lat)
        ]
    return loc_options, lat_options, lon_options, val_options


def build_candlestick_options(
    df, cs_x=None, cs_open=None, cs_close=None, cs_high=None, cs_low=None
):
    dtypes = get_dtypes(df)
    cols = sorted(dtypes.keys())
    num_cols, x_cols = [], []
    for c in cols:
        dtype = dtypes[c]
        classification = classify_type(dtype)
        if classification in ["S", "D"]:
            x_cols.append(c)
        if classification in ["F", "I"]:
            num_cols.append(c)

    x_options = [build_option(c) for c in x_cols]
    close_options = [
        build_option(c)
        for c in num_cols
        if c not in build_selections(cs_x, cs_open, cs_low, cs_high)
    ]
    open_options = [
        build_option(c)
        for c in num_cols
        if c not in build_selections(cs_x, cs_close, cs_low, cs_high)
    ]
    low_options = [
        build_option(c)
        for c in num_cols
        if c not in build_selections(cs_x, cs_open, cs_close, cs_high)
    ]
    high_options = [
        build_option(c)
        for c in num_cols
        if c not in build_selections(cs_x, cs_open, cs_close, cs_low)
    ]
    return x_options, close_options, open_options, low_options, high_options


def build_label_value_options(
    df, selected_value=None, selected_label=None, all_value=False
):
    dtypes = get_dtypes(df)
    cols = sorted(dtypes.keys())
    num_cols = []
    for c in cols:
        dtype = dtypes[c]
        classification = classify_type(dtype)
        if classification in ["F", "I"]:
            num_cols.append(c)

    value_options = [
        build_option(c) for c in num_cols if c not in build_selections(selected_label)
    ]
    if all_value:
        value_options = [build_option("_all_columns_", "All Columns")] + value_options
    label_options = [
        build_option(c) for c in cols if c not in build_selections(selected_value)
    ]
    return value_options, label_options


def build_label_value_store(prop, inputs):
    return dcc.Store(
        id="{}-input-data".format(prop),
        data={
            k: v
            for k, v in inputs.items()
            if k
            in [
                "{}_value".format(prop),
                "{}_label".format(prop),
                "{}_group".format(prop),
            ]
        },
    )


def bootstrap_checkbox_prop():
    if parse_version(dbc.__version__) >= parse_version("1.0.0"):
        return "value"
    return "checked"


def build_dropna(dropna, prop=None):
    if PY3:
        checkbox_kwargs = dict(
            id=(
                "{}-dropna-checkbox".format(prop)
                if prop is not None
                else "dropna-checkbox"
            ),
            style=dict(width="inherit"),
        )
        checkbox_kwargs[bootstrap_checkbox_prop()] = True if dropna is None else dropna
        return build_input(
            text("Dropna"),
            html.Div(dbc.Checkbox(**checkbox_kwargs), className="checkbox-wrapper"),
            className="col-auto",
            id="{}-dropna-input".format(prop) if prop is not None else "dropna-input",
        )
    return build_input(
        text("Dropna"),
        dcc.Input(
            id=(
                "{}-dropna-checkbox".format(prop)
                if prop is not None
                else "dropna-checkbox"
            ),
            type="hidden",
            value=True if dropna is None else dropna,
        ),
        id="{}-dropna-input".format(prop) if prop is not None else "dropna-input",
        style={"display": "none"},
    )


def build_label_value_inputs(
    prop,
    inputs,
    df,
    group_options,
    additional_inputs=[],
    multi_value=False,
    all_option=False,
):
    show_inputs = inputs.get("chart_type") == prop
    props = ["{}_value", "{}_label", "{}_group", "{}_dropna"]
    selected_value, selected_label, selected_group, dropna = (
        inputs.get(p.format(prop)) for p in props
    )
    all_options = build_label_value_options(
        df,
        selected_value=selected_value,
        selected_label=selected_label,
        all_value=multi_value and all_option,
    )
    value_options, label_options = all_options
    return html.Div(
        [
            build_input(
                text("Value(s)" if multi_value else "Value"),
                dcc.Dropdown(
                    id="{}-value-dropdown".format(prop),
                    options=value_options,
                    placeholder=text(
                        "Select Column(s)" if multi_value else "Select Column"
                    ),
                    style=dict(width="inherit"),
                    value=selected_value,
                    multi=multi_value,
                ),
            ),
            build_input(
                text("Labels"),
                dcc.Dropdown(
                    id="{}-label-dropdown".format(prop),
                    options=label_options,
                    placeholder=text("Select Column"),
                    style=dict(width="inherit"),
                    value=selected_label,
                ),
            ),
        ]
        + additional_inputs
        + [
            build_input(
                text("Group"),
                dcc.Dropdown(
                    id="{}-group-dropdown".format(prop),
                    options=group_options,
                    multi=True,
                    placeholder=text("Select Group(s)"),
                    value=selected_group,
                    style=dict(width="inherit"),
                ),
                className="col",
                id="{}-group-input".format(prop),
            ),
            build_dropna(dropna, prop),
        ],
        id="{}-inputs".format(prop),
        className="row charts-filters",
        style={} if show_inputs else {"display": "none"},
    )


def build_funnel_inputs(inputs, df, group_options):
    stacked_toggle = build_input(
        "{}?".format(text("Stack")),
        html.Div(
            daq.BooleanSwitch(id="funnel-stack-toggle", on=False),
            className="toggle-wrapper",
        ),
        id="funnel-stack-input",
        style=show_style(inputs.get("funnel_group") is not None),
        className="col-auto",
    )

    return build_label_value_inputs(
        "funnel", inputs, df, group_options, additional_inputs=[stacked_toggle]
    )


def get_num_cols(df):
    dtypes = get_dtypes(df)
    cols = sorted(dtypes.keys())
    num_cols = []
    for c in cols:
        dtype = dtypes[c]
        classification = classify_type(dtype)
        if classification in ["F", "I"]:
            num_cols.append(c)
    return num_cols


def build_pareto_options(df, x=None, bars=None, line=None):
    dtypes = get_dtypes(df)
    cols = sorted(dtypes.keys())
    num_cols = get_num_cols(df)

    x_options = [build_option(c) for c in cols if c not in [bars, line]]
    bars_options = [build_option(c) for c in num_cols if c not in [x, line]]
    line_options = [build_option(c) for c in num_cols if c not in [x, bars]]
    sort_options = [build_option(c) for c in cols]
    return x_options, bars_options, line_options, sort_options


def build_pareto_inputs(inputs, df, group_options):
    show_inputs = inputs.get("chart_type") == "pareto"
    x, bars, line, sort, sort_dir, group, dropna = (
        inputs.get("pareto_{}".format(prop))
        for prop in ["x", "bars", "line", "sort", "dir", "group", "dropna"]
    )
    x_options, bar_options, line_options, sort_options = build_pareto_options(
        df, x, bars, line
    )

    return html.Div(
        [
            build_input(
                [html.Div(text("X")), html.Small("({})".format(text("Agg By")))],
                dcc.Dropdown(
                    id="pareto-x-dropdown",
                    options=x_options,
                    placeholder="{} (1, 2, ..., N)".format(text("Default Index")),
                    value=x,
                    style=dict(width="inherit"),
                ),
                label_class="input-group-addon d-block pt-1 pb-0",
            ),
            build_input(
                text("Bars"),
                dcc.Dropdown(
                    id="pareto-bars-dropdown",
                    options=bar_options,
                    placeholder=text("Select Column"),
                    style=dict(width="inherit"),
                    value=bars,
                ),
            ),
            build_input(
                text("Line"),
                dcc.Dropdown(
                    id="pareto-line-dropdown",
                    options=line_options,
                    placeholder=text("Select Column"),
                    style=dict(width="inherit"),
                    value=line,
                ),
            ),
            build_input(
                text("Sort"),
                dcc.Dropdown(
                    id="pareto-sort-dropdown",
                    options=sort_options,
                    placeholder=text("Select Column"),
                    style=dict(width="inherit"),
                    value=sort,
                ),
            ),
            build_input(
                text("Dir"),
                dcc.Dropdown(
                    id="pareto-dir-dropdown",
                    options=[build_option(d, d.capitalize()) for d in ["ASC", "DESC"]],
                    style=dict(width="inherit"),
                    value=sort_dir or "DESC",
                    placeholder=None,
                    clearable=False,
                ),
            ),
            build_input(
                text("Group"),
                dcc.Dropdown(
                    id="pareto-group-dropdown",
                    options=group_options,
                    multi=True,
                    placeholder=text("Select Group(s)"),
                    value=group,
                    style=dict(width="inherit"),
                ),
                className="col",
                id="pareto-group-input",
            ),
            build_dropna(dropna, "pareto"),
        ],
        id="pareto-inputs",
        style={} if show_inputs else {"display": "none"},
        className="row p-0 charts-filters",
    )


def build_histogram_inputs(inputs, df, group_options):
    show_inputs = inputs.get("chart_type") == "histogram"
    col, histogram_type, bins, group = (
        inputs.get("histogram_{}".format(prop))
        for prop in ["col", "type", "bins", "group"]
    )
    col_options = get_num_cols(df)

    return html.Div(
        [
            build_input(
                text("Col"),
                dcc.Dropdown(
                    id="histogram-col-dropdown",
                    options=col_options,
                    value=col,
                    style=dict(width="inherit"),
                ),
            ),
            dcc.Tabs(
                id="histogram-type-tabs",
                value=histogram_type or "bins",
                children=[
                    build_tab(text("Bins"), "bins"),
                    build_tab(text("Density"), "density"),
                ],
                style=dict(height="36px", width="10em"),
            ),
            build_input(
                text("Bins"),
                dcc.Input(
                    id="histogram-bins-input",
                    type="number",
                    placeholder=text("Enter Bins"),
                    className="form-control text-center",
                    style={"lineHeight": "inherit"},
                    value=bins or 5,
                ),
                id="histogram-bins-div",
                className="col-md-1 pr-0",
                style={} if histogram_type == "bins" else {"display": "none"},
            ),
            build_input(
                text("Group"),
                dcc.Dropdown(
                    id="histogram-group-dropdown",
                    options=group_options,
                    multi=True,
                    placeholder=text("Select Group(s)"),
                    value=group,
                    style=dict(width="inherit"),
                ),
                className="col",
                id="histogram-group-input",
            ),
        ],
        id="histogram-inputs",
        style={} if show_inputs else {"display": "none"},
        className="row p-0 charts-filters",
    )


def build_mapbox_style_options():
    from dtale.charts.utils import get_mapbox_token

    free_styles = [
        "open-street-map",
        "carto-positron",
        "carto-darkmatter",
        "stamen-terrain",
        "stamen-toner",
        "stamen-watercolor",
    ]
    token_styles = [
        "basic",
        "streets",
        "outdoors",
        "light",
        "dark",
        "satellite",
        "satellite-streets",
    ]
    styles = free_styles
    if get_mapbox_token() is not None:
        styles += token_styles
    return [build_option(v) for v in styles]


def bar_input_style(**inputs):
    """
    Sets display CSS property for bar chart inputs
    """
    chart_type, group_col = (inputs.get(p) for p in ["chart_type", "group"])
    show_bar = chart_type == "bar"
    show_barsort = show_bar and group_col is None
    return (
        dict(display="block" if show_bar else "none"),
        dict(display="block" if show_barsort else "none"),
    )


def colorscale_input_style(**inputs):
    return dict(
        display=(
            "block"
            if inputs.get("chart_type")
            in ["heatmap", "maps", "3d_scatter", "surface", "clustergram"]
            else "none"
        )
    )


def animate_styles(df, **inputs):
    chart_type, agg, cpg, cpy = (
        inputs.get(p) for p in ["chart_type", "agg", "cpg", "cpy"]
    )
    opts = []
    if cpg or cpy or agg in ["pctsum", "pctct"]:
        return dict(display="none"), dict(display="none"), opts
    if chart_type in ANIMATION_CHARTS:
        return dict(display="block"), dict(display="none"), opts
    if chart_type in ANIMATE_BY_CHARTS:
        opts = [build_option(v, l) for v, l in build_cols(df.columns, get_dtypes(df))]
    if len(opts):
        return dict(display="none"), dict(display="block"), opts
    return dict(display="none"), dict(display="none"), []


def lock_zoom_style(chart_type):
    return (
        dict(display="block", textTransform="none")
        if chart_type in ["3d_scatter", "surface"]
        else dict(display="none")
    )


def show_chart_per_group(**inputs):
    """
    Boolean function to determine whether "Chart Per Group" toggle should be displayed or not
    """
    chart_type, group = (inputs.get(p) for p in ["chart_type", "group"])
    invalid_type = chart_type in ["pie", "wordcloud", "maps", "funnel", "clustergram"]
    return (
        show_input_handler(chart_type)("group")
        and len(group or [])
        and not invalid_type
    )


def show_chart_per_y(**inputs):
    """
    Boolean function to determine whether "Chart Per Y-Axis" toggle should be displayed or not
    """
    chart_type, y = (inputs.get(p) for p in ["chart_type", "y"])
    return show_input_handler(chart_type)("y", "multi") and len(y or []) > 1


def show_yaxis_ranges(**inputs):
    """
    Boolean function to determine whether "Y-Axis Range" inputs should be displayed or not
    """
    chart_type, y = (inputs.get(p) for p in ["chart_type", "y"])
    return chart_type in YAXIS_CHARTS and len(y or [])


def get_yaxis_type_tabs(y):
    tabs = [
        build_tab(text("Default"), "default", {"padding": "2px", "minWidth": "4em"}),
        build_tab(text("Single"), "single", {"padding": "2px", "minWidth": "4em"}),
    ]
    if len(y) <= 1:
        return tabs
    return tabs + [
        build_tab(text("Multi"), "multi", {"padding": "2px", "minWidth": "4em"})
    ]


def get_yaxis_scale_tabs():
    return [
        build_tab(text("Linear"), "linear", {"padding": "2px", "minWidth": "4em"}),
        build_tab(text("Log"), "log", {"padding": "2px", "minWidth": "4em"}),
    ]


def build_group_val_options(df, group_cols):
    group_vals = find_group_vals(df, group_cols)
    return [
        build_option(
            json.dumps(gv),
            CHART_JOINER_CHAR.join([str(gv.get(p, "NaN")) for p in group_cols]),
        )
        for gv in group_vals
    ]


def build_map_type_tabs(map_type):
    def _build_map_type_hoverable():
        for t in MAP_TYPES:
            if t.get("image", False):
                yield html.Div(
                    [
                        html.Span(text(t.get("label", t["value"].capitalize()))),
                        html.Img(src=build_img_src(t["value"], img_type="map_type")),
                    ],
                    className="col-md-4",
                )

    return html.Div(
        [
            dcc.Tabs(
                id="map-type-tabs",
                value=map_type or "choropleth",
                children=[
                    build_tab(text(t.get("label", t["value"].capitalize())), t["value"])
                    for t in MAP_TYPES
                ],
                style=dict(height="36px"),
            ),
            html.Div(
                html.Div(list(_build_map_type_hoverable()), className="row"),
                className="hoverable__content map-types",
            ),
        ],
        style=dict(paddingLeft=15, borderBottom="none", width="20em"),
        className="hoverable",
    )


def main_inputs_and_group_val_display(inputs):
    group_types = get_group_types(inputs)
    if len(group_types):
        is_groups = ["groups"] == group_types or (
            "groups" in group_types and inputs.get("group_type") == "groups"
        )
        is_bins = ["bins"] == group_types or (
            "bins" in group_types and inputs.get("group_type") == "bins"
        )
        return (
            dict(display="block" if len(group_types) > 1 else "none"),
            dict(display="block" if is_groups else "none"),
            dict(display="block" if is_bins else "none"),
            "col-md-8",
            dict(display="block"),
        )
    return (
        dict(display="none"),
        dict(display="none"),
        dict(display="none"),
        "col-md-12",
        dict(display="none"),
    )


def build_slider_counts(df, data_id, query_value):
    record_ct = len(
        run_query(
            handle_predefined(data_id, df=df),
            build_query(data_id, query_value),
            global_state.get_context_variables(data_id),
        )
    )
    slider_counts = {
        "{}".format(v * 20): {
            "label": "{}% ({:,.0f})".format(v * 20, (v * 2) / 10 * record_ct)
        }
        for v in range(1, 6)
    }
    slider_counts["100"]["style"] = {"white-space": "nowrap"}
    return slider_counts


def collapse_btn_text(is_open, label):
    return "{} {}".format("\u25bc" if is_open else "\u25b6", label)


def charts_layout(df, settings, **inputs):
    """
    Builds main dash inputs with dropdown options populated with the columns of the dataframe associated with the
    page. Inputs included are: chart tabs, query, x, y, z, group, aggregation, rolling window/computation,
    chart per group toggle, bar sort, bar mode, y-axis range editors

    :param df: dataframe to drive the charts built on page
    :type df: :class:`pandas:pandas.DataFrame`
    :param settings: global settings associated with this dataframe (contains properties like "query")
    :type param: dict
    :return: dash markup
    """
    chart_type, x, y, z, group, dropna, agg, load, load_type, stratified_group = (
        inputs.get(p)
        for p in [
            "chart_type",
            "x",
            "y",
            "z",
            "group",
            "dropna",
            "agg",
            "load",
            "load_type",
            "stratified_group",
        ]
    )
    loc_modes = loc_mode_info()
    y = y or []
    show_input = show_input_handler(chart_type)
    show_cpg = show_chart_per_group(**inputs)
    show_cpy = show_chart_per_y(**inputs)
    show_yaxis = show_yaxis_ranges(**inputs)
    scatter_input = dict(display="block" if chart_type == "scatter" else "none")
    bar_style, barsort_input_style = bar_input_style(**inputs)
    animate_style, animate_by_style, animate_opts = animate_styles(df, **inputs)
    custom_filtering = global_state.load_flag(
        inputs["data_id"], "enable_custom_filters", False
    )

    options = build_input_options(df, **inputs)
    (
        x_options,
        y_multi_options,
        y_single_options,
        z_options,
        group_options,
        barsort_options,
        yaxis_options,
        stratified_groups,
    ) = options
    query_placeholder = "{} (ex: col1 == 1)".format(text("Enter pandas query"))
    query_value = inputs.get("query") or inner_build_query(
        settings, settings.get("query")
    )

    query_label = html.Div(
        [
            html.Span(text("Query")),
            html.A(
                html.I(className="fa fa-info-circle ml-4"),
                href="https://pandas.pydata.org/pandas-docs/stable/user_guide/indexing.html#indexing-query",
                target="_blank",
                style={"color": "white"},
            ),
        ],
        className="input-group-addon",
        style={"minWidth": "7em"},
    )
    yaxis_type = (inputs.get("yaxis") or {}).get("type") or "default"
    yaxis_type_style = (
        {"borderRadius": "0 0.25rem 0.25rem 0"} if yaxis_type == "default" else None
    )
    show_map = chart_type == "maps"
    map_props = ["map_type", "loc_mode", "loc", "lat", "lon", "map_val"]
    map_type, loc_mode, loc, lat, lon, map_val = (inputs.get(p) for p in map_props)
    map_scope, proj, mapbox_style = (
        inputs.get(p) for p in ["scope", "proj", "mapbox_style"]
    )
    loc_options, lat_options, lon_options, map_val_options = build_map_options(
        df, type=map_type, loc=loc, lat=lat, lon=lon, map_val=map_val
    )
    show_candlestick = chart_type == "candlestick"
    cs_props = [
        "cs_x",
        "cs_open",
        "cs_close",
        "cs_high",
        "cs_low",
        "cs_group",
        "cs_dropna",
    ]
    cs_x, cs_open, cs_close, cs_high, cs_low, cs_group, cs_dropna = (
        inputs.get(p) for p in cs_props
    )
    (
        cs_x_options,
        open_options,
        close_options,
        high_options,
        low_options,
    ) = build_candlestick_options(
        df,
        cs_x=cs_x,
        cs_open=cs_open,
        cs_close=cs_close,
        cs_high=cs_high,
        cs_low=cs_low,
    )
    cscale_style = colorscale_input_style(**inputs)
    default_cscale = DEFAULT_CSALES.get(chart_type, REDS)

    (
        group_type_style,
        group_val_style,
        bins_style,
        main_input_class,
        show_groups,
    ) = main_inputs_and_group_val_display(inputs)
    group_val = [json.dumps(gv) for gv in inputs.get("group_val") or []]

    def show_map_style(show):
        return {} if show else {"display": "none"}

    body_items = [
        dcc.Store(id="query-data", data=query_value),
        dcc.Store(
            id="input-data",
            data={
                k: v
                for k, v in inputs.items()
                if k not in ["cpg", "cpy", "barmode", "barsort"]
            },
        ),
        dcc.Store(
            id="chart-input-data",
            data={
                k: v
                for k, v in inputs.items()
                if k in ["cpg", "cpy", "barmode", "barsort"]
            },
        ),
        dcc.Store(
            id="map-input-data",
            data={
                k: v
                for k, v in inputs.items()
                if k
                in [
                    "map_type",
                    "map_code",
                    "loc_mode",
                    "lat",
                    "lon",
                    "map_val",
                    "scope",
                    "proj",
                ]
            },
        ),
        dcc.Store(
            id="candlestick-input-data",
            data={
                k: v
                for k, v in inputs.items()
                if k in ["cs_x", "cs_open", "cs_close", "cs_high", "cs_low", "cs_group"]
            },
        ),
        build_label_value_store("treemap", inputs),
        build_label_value_store("funnel", inputs),
        build_label_value_store("clustergram", inputs),
        dcc.Store(
            id="pareto-input-data",
            data={
                k: v
                for k, v in inputs.items()
                if k
                in [
                    "pareto_x",
                    "pareto_bars",
                    "pareto_line",
                    "pareto_sort",
                    "pareto_dir",
                    "pareto_group",
                ]
            },
        ),
        dcc.Store(
            id="histogram-input-data",
            data={
                k: v
                for k, v in inputs.items()
                if k
                in [
                    "histogram_col",
                    "histogram_type",
                    "histogram_bins",
                    "histogram_group",
                ]
            },
        ),
        dcc.Store(id="range-data"),
        dcc.Store(id="yaxis-data", data=inputs.get("yaxis")),
        dcc.Store(id="last-chart-input-data", data=inputs),
        dcc.Store(id="load-clicks", data=0),
        dcc.Store(id="save-clicks", data=0),
        dcc.Store(id="y-all-clicks", data=0),
        dcc.Store(id="z-all-clicks", data=0),
        dcc.Input(id="chart-code", type="hidden"),
        html.Div(
            [
                html.Div(
                    dbc.Button(
                        collapse_btn_text(False, text("Data Selection")),
                        id="collapse-data-btn",
                    ),
                    className="col-auto pr-0",
                ),
                html.Div(
                    dbc.Collapse(
                        dbc.Card(
                            dbc.CardBody(
                                [
                                    dcc.Tabs(
                                        id="data-tabs",
                                        value=inputs["data_id"],
                                        children=[
                                            build_tab(
                                                global_state.get_name(k) or k,
                                                str(k),
                                                id="data-tab-{}".format(k),
                                            )
                                            for k in global_state.keys()
                                        ],
                                        style=dict(height="36px"),
                                    )
                                ]
                                + [
                                    dbc.Tooltip(
                                        global_state.get_name(k) or k,
                                        target="data-tab-{}".format(k),
                                    )
                                    for k in global_state.keys()
                                ]
                            )
                        ),
                        id="collapse-data",
                    ),
                    className="col",
                ),
            ],
            className="row pb-3 charts-filters",
            style=dict(display="none" if len(global_state.keys()) < 2 else "block"),
        ),
        html.Div(
            html.Div(
                [
                    dcc.Tabs(
                        id="chart-tabs",
                        value=chart_type or "line",
                        children=[
                            build_tab(
                                text(t.get("label", t["value"].capitalize())),
                                t["value"],
                                id="chart-type-tab-{}".format(i),
                            )
                            for i, t in enumerate(CHARTS)
                        ],
                        style=dict(height="36px"),
                    )
                ]
                + [
                    dbc.Tooltip(
                        text(t.get("label", t["value"].capitalize())),
                        target="chart-type-tab-{}".format(i),
                    )
                    for i, t in enumerate(CHARTS)
                ],
                className="col-md-12",
            ),
            className="row pt-3 pb-3 charts-filters",
        ),
        (
            html.Div(
                html.Div(
                    [
                        html.Div(
                            [
                                query_label,
                                dcc.Input(
                                    id="query-input",
                                    type="text",
                                    placeholder=query_placeholder,
                                    className="form-control",
                                    value=query_value,
                                    style={
                                        "lineHeight": "inherit",
                                        "display": (
                                            "inherit" if custom_filtering else "none"
                                        ),
                                    },
                                    disabled=not custom_filtering,
                                ),
                            ],
                            className="input-group mr-3",
                        )
                    ],
                    className="col",
                ),
                className="row pt-3 pb-3 charts-filters",
            )
        ),
        html.Div(
            html.Div(
                [
                    html.Div(
                        [
                            build_hoverable(
                                html.Span(text("Load")),
                                load_type_msg(),
                                additional_classes="load-types-tooltip mb-auto mt-auto",
                                top="100%",
                            ),
                            dcc.Dropdown(
                                id="load-type-dropdown",
                                options=[
                                    build_option(v, v.capitalize()) for v in LOAD_TYPES
                                ],
                                className="pl-5",
                                value=load_type or "random",
                                clearable=False,
                            ),
                            dcc.Dropdown(
                                id="stratified-group-dropdown",
                                options=stratified_groups,
                                value=stratified_group,
                                clearable=False,
                                style=(
                                    {}
                                    if load_type == "stratified"
                                    else {"display": "none"}
                                ),
                            ),
                            dcc.Slider(
                                id="load-input",
                                min=10,
                                max=100,
                                step=10,
                                value=100 if load is None else load,
                                className="w-100",
                                marks=build_slider_counts(
                                    df, inputs["data_id"], query_value
                                ),
                                tooltip={"always_visible": False, "placement": "left"},
                            ),
                        ],
                        className="input-group mr-3",
                    )
                ],
                className="col",
            ),
            className="row pt-3 pb-0 charts-filters",
        ),
        html.Div(
            [
                html.Div(
                    [
                        dbc.Button(
                            collapse_btn_text(False, text("Cleaners")),
                            id="collapse-cleaners-btn",
                        ),
                        html.Span(
                            "",
                            id="selected-cleaners",
                            className="pl-3",
                            style=dict(fontSize="85%"),
                        ),
                    ],
                    className="col-auto pr-0",
                ),
                html.Div(
                    dbc.Collapse(
                        dbc.Card(
                            dbc.CardBody(
                                dcc.Dropdown(
                                    id="cleaners-dropdown",
                                    options=[
                                        build_option(c["value"], c["label"])
                                        for c in get_cleaner_configs()
                                    ],
                                    multi=True,
                                    placeholder=text("Select Cleaner(s)"),
                                    value=group,
                                    style=dict(width="inherit"),
                                )
                            )
                        ),
                        id="collapse-cleaners",
                    ),
                    className="col",
                ),
            ],
            className="row pb-3 charts-filters",
            style=dict(display="block"),
        ),
        html.Div(
            [
                html.Div(
                    [
                        html.Div(
                            [
                                build_input(
                                    [
                                        html.Div(text("X")),
                                        html.Small("({})".format(text("Agg By"))),
                                    ],
                                    dcc.Dropdown(
                                        id="x-dropdown",
                                        options=x_options,
                                        placeholder="{} (1, 2, ..., N)".format(
                                            text("Default Index")
                                        ),
                                        value=x,
                                        style=dict(width="inherit"),
                                    ),
                                    label_class="input-group-addon d-block pt-1 pb-0",
                                    style=show_style(show_input("x")),
                                ),
                                build_input(
                                    text("Y"),
                                    [
                                        dcc.Dropdown(
                                            id="y-multi-dropdown",
                                            options=y_multi_options,
                                            multi=True,
                                            placeholder=text("Select Column(s)"),
                                            style=dict(width="inherit"),
                                            value=(
                                                y if show_input("y", "multi") else None
                                            ),
                                        ),
                                        dbc.Button(
                                            html.I(
                                                className="fa-solid fa-check-double"
                                            ),
                                            id="y-select-all-btn",
                                            className="ml-3",
                                            color="secondary",
                                            style=dict(
                                                height="24px",
                                                padding=".25rem .2rem",
                                                marginTop="auto",
                                                marginBottom="auto",
                                            ),
                                        ),
                                        dbc.Tooltip(
                                            "Select All Y", target="y-select-all-btn"
                                        ),
                                    ],
                                    className="col",
                                    id="y-multi-input",
                                    style=show_style(show_input("y", "multi")),
                                ),
                                build_input(
                                    [
                                        html.Div(text("Y")),
                                        html.Small("({})".format(text("Agg By"))),
                                    ],
                                    dcc.Dropdown(
                                        id="y-single-dropdown",
                                        options=y_single_options,
                                        placeholder=text("Select Column"),
                                        style=dict(width="inherit"),
                                        value=(
                                            y[0] if show_input("y") and len(y) else None
                                        ),
                                    ),
                                    className="col",
                                    label_class="input-group-addon d-block pt-1 pb-0",
                                    id="y-single-input",
                                    style=show_style(show_input("y")),
                                ),
                                build_input(
                                    text("Z"),
                                    dcc.Dropdown(
                                        id="z-dropdown",
                                        options=z_options,
                                        placeholder=text("Select Column"),
                                        style=dict(width="inherit"),
                                        value=z,
                                    ),
                                    className="col",
                                    id="z-input",
                                    style=show_style(show_input("z")),
                                ),
                                build_input(
                                    text("Group"),
                                    dcc.Dropdown(
                                        id="group-dropdown",
                                        options=group_options,
                                        multi=True,
                                        placeholder=text("Select Group(s)"),
                                        value=group,
                                        style=dict(width="inherit"),
                                    ),
                                    className="col",
                                    id="group-input",
                                    style=show_style(show_input("group")),
                                ),
                                build_dropna(dropna),
                            ],
                            id="standard-inputs",
                            style=(
                                {}
                                if not show_map
                                and not show_candlestick
                                and chart_type
                                not in ["treemap", "funnel", "clustergram"]
                                else {"display": "none"}
                            ),
                            className="row p-0 charts-filters",
                        ),
                        html.Div(
                            [
                                build_map_type_tabs(map_type),
                                html.Div(
                                    [
                                        html.Div(
                                            [
                                                build_loc_mode_hover(loc_mode),
                                                dcc.Dropdown(
                                                    id="map-loc-mode-dropdown",
                                                    options=[
                                                        build_option(
                                                            v, loc_modes[v].get("label")
                                                        )
                                                        for v in [
                                                            "ISO-3",
                                                            "USA-states",
                                                            "country names",
                                                            "geojson-id",
                                                        ]
                                                    ],
                                                    style=dict(width="inherit"),
                                                    value=loc_mode,
                                                ),
                                            ],
                                            className="input-group mr-3",
                                        )
                                    ],
                                    id="map-loc-mode-input",
                                    style=show_map_style(map_type == "choropleth"),
                                    className="col-auto",
                                ),
                                custom_geojson.build_modal(map_type, loc_mode),
                                build_input(
                                    [
                                        html.Div(text("Locations")),
                                        html.Small("({})".format(text("Agg By"))),
                                    ],
                                    dcc.Dropdown(
                                        id="map-loc-dropdown",
                                        options=loc_options,
                                        placeholder=text("Select Column"),
                                        value=loc,
                                        style=dict(width="inherit"),
                                    ),
                                    id="map-loc-input",
                                    label_class="input-group-addon d-block pt-1 pb-0",
                                    style=show_map_style(map_type == "choropleth"),
                                ),
                                build_input(
                                    [
                                        html.Div(text("Lat")),
                                        html.Small("({})".format(text("Agg By"))),
                                    ],
                                    dcc.Dropdown(
                                        id="map-lat-dropdown",
                                        options=lat_options,
                                        placeholder="Select a column",
                                        value=lat,
                                        style=dict(width="inherit"),
                                    ),
                                    id="map-lat-input",
                                    label_class="input-group-addon d-block pt-1 pb-0",
                                    style=show_map_style(
                                        map_type in ["scattergeo", "mapbox"]
                                    ),
                                ),
                                build_input(
                                    [
                                        html.Div(text("Lon")),
                                        html.Small("({})".format(text("Agg By"))),
                                    ],
                                    dcc.Dropdown(
                                        id="map-lon-dropdown",
                                        options=lon_options,
                                        placeholder="Select a column",
                                        style=dict(width="inherit"),
                                        value=lon,
                                    ),
                                    id="map-lon-input",
                                    label_class="input-group-addon d-block pt-1 pb-0",
                                    style=show_map_style(
                                        map_type in ["scattergeo", "mapbox"]
                                    ),
                                ),
                                build_input(
                                    text("Scope"),
                                    dcc.Dropdown(
                                        id="map-scope-dropdown",
                                        options=[build_option(v) for v in SCOPES],
                                        style=dict(width="inherit"),
                                        value=map_scope or "world",
                                    ),
                                    id="map-scope-input",
                                    style=show_map_style(map_type == "scattergeo"),
                                ),
                                html.Div(
                                    [
                                        html.Div(
                                            [
                                                build_mapbox_token_hover(),
                                                dcc.Dropdown(
                                                    id="map-mapbox-style-dropdown",
                                                    options=build_mapbox_style_options(),
                                                    style=dict(width="inherit"),
                                                    value=mapbox_style
                                                    or "open-street-map",
                                                ),
                                            ],
                                            className="input-group mr-3",
                                        )
                                    ],
                                    id="map-mapbox-style-input",
                                    className="col-auto",
                                    style=show_map_style(map_type == "mapbox"),
                                ),
                                html.Div(
                                    [
                                        html.Div(
                                            [
                                                build_proj_hover(proj),
                                                dcc.Dropdown(
                                                    id="map-proj-dropdown",
                                                    options=[
                                                        build_option(v)
                                                        for v in PROJECTIONS
                                                    ],
                                                    style=dict(width="inherit"),
                                                    value=proj,
                                                ),
                                            ],
                                            className="input-group mr-3",
                                        )
                                    ],
                                    id="map-proj-input",
                                    style=show_map_style(map_type == "scattergeo"),
                                    className="col-auto",
                                ),
                                build_input(
                                    text("Value"),
                                    dcc.Dropdown(
                                        id="map-val-dropdown",
                                        options=map_val_options,
                                        placeholder=text("Select Column"),
                                        style=dict(width="inherit"),
                                        value=map_val,
                                    ),
                                ),
                                build_input(
                                    text("Group"),
                                    dcc.Dropdown(
                                        id="map-group-dropdown",
                                        options=group_options,
                                        multi=True,
                                        placeholder=text("Select Group(s)"),
                                        value=inputs.get("map_group"),
                                        style=dict(width="inherit"),
                                    ),
                                    className="col",
                                    id="map-group-input",
                                ),
                                build_dropna(inputs.get("map_dropna", True), "map"),
                            ],
                            id="map-inputs",
                            className="row charts-filters",
                            style={} if show_map else {"display": "none"},
                        ),
                        html.Div(
                            [
                                build_input(
                                    text("X"),
                                    dcc.Dropdown(
                                        id="candlestick-x-dropdown",
                                        options=cs_x_options,
                                        placeholder=text("Select Column"),
                                        style=dict(width="inherit"),
                                        value=cs_x,
                                    ),
                                ),
                                build_input(
                                    text("Open"),
                                    dcc.Dropdown(
                                        id="candlestick-open-dropdown",
                                        options=open_options,
                                        placeholder=text("Select Column"),
                                        style=dict(width="inherit"),
                                        value=cs_open,
                                    ),
                                ),
                                build_input(
                                    text("High"),
                                    dcc.Dropdown(
                                        id="candlestick-high-dropdown",
                                        options=high_options,
                                        placeholder=text("Select Column"),
                                        style=dict(width="inherit"),
                                        value=cs_high,
                                    ),
                                ),
                                build_input(
                                    text("Low"),
                                    dcc.Dropdown(
                                        id="candlestick-low-dropdown",
                                        options=low_options,
                                        placeholder=text("Select Column"),
                                        style=dict(width="inherit"),
                                        value=cs_low,
                                    ),
                                ),
                                build_input(
                                    text("Close"),
                                    dcc.Dropdown(
                                        id="candlestick-close-dropdown",
                                        options=close_options,
                                        placeholder=text("Select Column"),
                                        style=dict(width="inherit"),
                                        value=cs_close,
                                    ),
                                ),
                                build_input(
                                    text("Group"),
                                    dcc.Dropdown(
                                        id="candlestick-group-dropdown",
                                        options=group_options,
                                        multi=True,
                                        placeholder=text("Select Group(s)"),
                                        value=inputs.get("cs_group"),
                                        style=dict(width="inherit"),
                                    ),
                                    className="col",
                                    id="candlestick-group-input",
                                ),
                                build_dropna(
                                    inputs.get("cs_dropna", True), "candlestick"
                                ),
                            ],
                            id="candlestick-inputs",
                            className="row charts-filters",
                            style={} if show_candlestick else {"display": "none"},
                        ),
                        build_label_value_inputs("treemap", inputs, df, group_options),
                        build_funnel_inputs(inputs, df, group_options),
                        build_label_value_inputs(
                            "clustergram",
                            inputs,
                            df,
                            group_options,
                            multi_value=True,
                            all_option=True,
                        ),
                        build_pareto_inputs(inputs, df, group_options),
                        build_histogram_inputs(inputs, df, group_options),
                        html.Div(
                            [
                                html.Div(
                                    [
                                        build_input(
                                            text("Aggregation"),
                                            dcc.Dropdown(
                                                id="agg-dropdown",
                                                options=[
                                                    build_option(v, text(AGGS[v]))
                                                    for v in [
                                                        "count",
                                                        "pctct",
                                                        "nunique",
                                                        "sum",
                                                        "mean",
                                                        "rolling",
                                                        "corr",
                                                        "first",
                                                        "last",
                                                        "drop_duplicates",
                                                        "median",
                                                        "min",
                                                        "max",
                                                        "std",
                                                        "var",
                                                        "mad",
                                                        "prod",
                                                        "pctsum",
                                                    ]
                                                ],
                                                placeholder=text("No Aggregation"),
                                                style=dict(width="inherit"),
                                                value=agg or "raw",
                                                disabled=len(
                                                    inputs.get(
                                                        "extended_aggregation", []
                                                    )
                                                )
                                                > 0,
                                            ),
                                        ),
                                        html.Span(
                                            "* {}".format(
                                                text(
                                                    "Extended Aggregation is currently populated."
                                                )
                                            ),
                                            id="ext-agg-warning",
                                            style=show_style(
                                                len(
                                                    inputs.get(
                                                        "extended_aggreation", []
                                                    )
                                                )
                                                > 0
                                            ),
                                            className="ext-agg-warning",
                                        ),
                                    ]
                                )
                            ]
                            + extended_aggregations.build_modal(
                                inputs.get("extended_aggregation", []), chart_type, y
                            )
                            + [
                                html.Div(
                                    [
                                        build_input(
                                            text("Window"),
                                            dcc.Input(
                                                id="window-input",
                                                type="number",
                                                placeholder=text("Enter Days"),
                                                className="form-control text-center",
                                                style={"lineHeight": "inherit"},
                                                value=inputs.get("window"),
                                            ),
                                        ),
                                        build_input(
                                            text("Computation"),
                                            dcc.Dropdown(
                                                id="rolling-comp-dropdown",
                                                options=[
                                                    build_option(
                                                        "corr", text("Correlation")
                                                    ),
                                                    build_option(
                                                        "count", text("Count")
                                                    ),
                                                    build_option(
                                                        "cov", text("Covariance")
                                                    ),
                                                    build_option(
                                                        "kurt", text("Kurtosis")
                                                    ),
                                                    build_option(
                                                        "max", text("Maximum")
                                                    ),
                                                    build_option("mean", text("Mean")),
                                                    build_option(
                                                        "median", text("Median")
                                                    ),
                                                    build_option(
                                                        "min", text("Minimum")
                                                    ),
                                                    build_option("skew", text("Skew")),
                                                    build_option(
                                                        "std",
                                                        text("Standard Deviation"),
                                                    ),
                                                    build_option("sum", text("Sum")),
                                                    build_option(
                                                        "var", text("Variance")
                                                    ),
                                                ],
                                                placeholder=text("Select Computation"),
                                                style=dict(width="inherit"),
                                                value=inputs.get("rolling_comp"),
                                            ),
                                        ),
                                    ],
                                    id="rolling-inputs",
                                    style=show_style(agg == "rolling"),
                                ),
                                build_input(
                                    text("Drilldowns"),
                                    html.Div(
                                        daq.BooleanSwitch(
                                            id="drilldown-toggle", on=False
                                        ),
                                        className="toggle-wrapper",
                                    ),
                                    id="drilldown-input",
                                    style=show_style((agg or "raw") != "raw"),
                                    className="col-auto",
                                ),
                            ],
                            className="row pt-3 pb-3 charts-filters",
                            id="charts-filters-div",
                            style=(
                                {"display": "none"} if chart_type == "histogram" else {}
                            ),
                        ),
                    ],
                    id="main-inputs",
                    className=main_input_class,
                ),
                html.Div(
                    [
                        html.Div(
                            html.Div(
                                html.Div(
                                    [
                                        html.Span(
                                            text("Group Type"),
                                            className="input-group-addon",
                                        ),
                                        html.Div(
                                            dcc.Tabs(
                                                id="group-type",
                                                value=inputs.get("group_type")
                                                or "groups",
                                                children=[
                                                    build_tab(
                                                        text("Values"),
                                                        "groups",
                                                        {
                                                            "padding": "2px",
                                                            "minWidth": "4em",
                                                        },
                                                    ),
                                                    build_tab(
                                                        text("Bins"),
                                                        "bins",
                                                        {
                                                            "padding": "2px",
                                                            "minWidth": "4em",
                                                        },
                                                    ),
                                                ],
                                            ),
                                            id="group-type-div",
                                            className="form-control col-auto pt-3",
                                        ),
                                    ],
                                    className="input-group",
                                ),
                                className="addon-min-width",
                            ),
                            className="col-md-12 p-0 pb-5",
                            id="group-type-input",
                            style=group_type_style,
                        ),
                        build_input(
                            text("Group(s)"),
                            dcc.Dropdown(
                                id="group-val-dropdown",
                                multi=True,
                                placeholder=text("Select Group Value(s)"),
                                value=group_val,
                                style=dict(width="inherit"),
                            ),
                            className="col-md-12 p-0 pb-5",
                            id="group-val-input",
                            style=group_val_style,
                        ),
                        build_input(
                            text("Bins"),
                            [
                                daq.NumericInput(
                                    id="bins-val-input",
                                    min=1,
                                    max=30,
                                    value=inputs.get("bins_val") or 5,
                                ),
                                html.Div(
                                    html.Div(
                                        [
                                            html.Span(
                                                text("Binning"),
                                                className="input-group-addon",
                                            ),
                                            html.Div(
                                                build_hoverable(
                                                    dcc.Tabs(
                                                        id="bin-type",
                                                        value=inputs.get("bin_type")
                                                        or "width",
                                                        children=[
                                                            build_tab(
                                                                text("Width"),
                                                                "width",
                                                                {
                                                                    "padding": "2px",
                                                                    "minWidth": "4em",
                                                                },
                                                            ),
                                                            build_tab(
                                                                text("Freq"),
                                                                "freq",
                                                                {
                                                                    "padding": "2px",
                                                                    "minWidth": "4em",
                                                                },
                                                            ),
                                                        ],
                                                    ),
                                                    bin_type_msg(),
                                                    "",
                                                    top="120%",
                                                ),
                                                id="bin-type-div",
                                                className="form-control col-auto pt-3",
                                            ),
                                        ],
                                        className="input-group",
                                    ),
                                    className="col-auto addon-min-width",
                                    id="bin-type-input",
                                ),
                            ],
                            className="col-md-12 p-0",
                            id="bins-input",
                            style=bins_style,
                        ),
                    ],
                    id="group-inputs-row",
                    className="col-md-4 row pt-3 pb-5",
                    style=show_groups,
                ),
            ],
            className="row",
        ),
        html.Div(
            [
                build_input(
                    text("Chart Per\nGroup"),
                    html.Div(
                        daq.BooleanSwitch(
                            id="cpg-toggle", on=inputs.get("cpg") or False
                        ),
                        className="toggle-wrapper",
                    ),
                    id="cpg-input",
                    style=show_style(show_cpg),
                    className="col-auto",
                ),
                build_input(
                    text("Chart Per\nY"),
                    html.Div(
                        daq.BooleanSwitch(
                            id="cpy-toggle", on=inputs.get("cpy") or False
                        ),
                        className="toggle-wrapper",
                    ),
                    id="cpy-input",
                    style=show_style(show_cpy),
                    className="col-auto",
                ),
                build_input(
                    text("Trendline"),
                    dcc.Dropdown(
                        id="trendline-dropdown",
                        options=[build_option("ols"), build_option("lowess")],
                        value=inputs.get("trendline"),
                    ),
                    className="col-auto addon-min-width",
                    style=scatter_input,
                    id="trendline-input",
                ),
                build_input(
                    text("Barmode"),
                    dcc.Dropdown(
                        id="barmode-dropdown",
                        options=[
                            build_option("group", text("Group")),
                            build_option("stack", text("Stack")),
                            build_option("relative", text("Relative")),
                        ],
                        value=inputs.get("barmode") or "group",
                        placeholder=text("Select Mode"),
                    ),
                    className="col-auto addon-min-width",
                    style=bar_style,
                    id="barmode-input",
                ),
                build_input(
                    text("Barsort"),
                    dcc.Dropdown(
                        id="barsort-dropdown",
                        options=barsort_options,
                        value=inputs.get("barsort"),
                    ),
                    className="col-auto addon-min-width",
                    style=barsort_input_style,
                    id="barsort-input",
                ),
                build_input(
                    text("Top"),
                    dcc.Dropdown(
                        id="top-bars",
                        options=[build_option(v) for v in [5, 10, 20, 50]],
                        value=inputs.get("top_bars"),
                        placeholder=None,
                    ),
                    className="col-auto",
                    style=barsort_input_style,
                    id="top-bars-input",
                ),
                html.Div(
                    html.Div(
                        [
                            html.Span(text("Y-Axis"), className="input-group-addon"),
                            html.Div(
                                [
                                    dcc.Tabs(
                                        id="yaxis-scale",
                                        value=inputs.get("scale") or "linear",
                                        children=get_yaxis_scale_tabs(),
                                        className="pr-5",
                                    ),
                                    dcc.Tabs(
                                        id="yaxis-type",
                                        value=yaxis_type,
                                        children=get_yaxis_type_tabs(y),
                                    ),
                                ],
                                id="yaxis-type-div",
                                className="form-control col-auto pt-3",
                                style=yaxis_type_style,
                            ),
                            dcc.Dropdown(id="yaxis-dropdown", options=yaxis_options),
                            html.Span(
                                "{}:".format(text("Min")),
                                className="input-group-addon col-auto",
                                id="yaxis-min-label",
                            ),
                            dcc.Input(
                                id="yaxis-min-input",
                                type="number",
                                className="form-control col-auto",
                                style={"lineHeight": "inherit"},
                            ),
                            html.Span(
                                "{}:".format(text("Max")),
                                className="input-group-addon col-auto",
                                id="yaxis-max-label",
                            ),
                            dcc.Input(
                                id="yaxis-max-input",
                                type="number",
                                className="form-control col-auto",
                                style={"lineHeight": "inherit"},
                            ),
                        ],
                        className="input-group",
                        id="yaxis-min-max-options",
                    ),
                    className="col-auto addon-min-width",
                    id="yaxis-input",
                    style=show_style(show_yaxis),
                ),
                build_input(
                    text("Colorscale"),
                    dash_components.DashColorscales(
                        id="colorscale-picker",
                        colorscale=inputs.get("colorscale") or default_cscale,
                    ),
                    className="col-auto addon-min-width pr-0",
                    style=cscale_style,
                    id="colorscale-input",
                ),
                build_input(
                    text("Animate"),
                    html.Div(
                        daq.BooleanSwitch(
                            id="animate-toggle", on=inputs.get("animate") or False
                        ),
                        className="toggle-wrapper",
                    ),
                    id="animate-input",
                    style=animate_style,
                    className="col-auto",
                ),
                build_input(
                    text("Animate By"),
                    dcc.Dropdown(
                        id="animate-by-dropdown",
                        options=animate_opts,
                        value=inputs.get("animate_by"),
                    ),
                    className="col-auto addon-min-width",
                    style=animate_by_style,
                    id="animate-by-input",
                ),
                dbc.Button(
                    text("Lock Zoom"),
                    id="lock-zoom-btn",
                    className="ml-auto",
                    style=lock_zoom_style(chart_type),
                ),
                build_input(
                    build_hoverable(
                        html.Div(text("Auto-Load"), style=dict(color="white")),
                        auto_load_msg(),
                        "",
                        top="120%",
                    ),
                    html.Div(
                        daq.BooleanSwitch(
                            id="auto-load-toggle", on=True, color="green"
                        ),
                        className="toggle-wrapper",
                    ),
                    id="auto-load-input",
                    className="ml-auto col-auto",
                ),
                dbc.Button(
                    text("Load"),
                    id="load-btn",
                    color="primary",
                    style=dict(display="none"),
                ),
                build_hoverable(
                    dbc.Button(
                        text("Save"),
                        id="save-btn",
                        color="primary",
                        style=dict(display="none"),
                    ),
                    text("save_msg"),
                    "",
                    top="120%",
                ),
                build_hoverable(
                    html.A(
                        html.I(className="fas fa-file-code fa-xl"),
                        id="export-all-chart-btn",
                        href="",
                        className="export-chart-btn",
                        style=dict(display="none"),
                    ),
                    "Export Charts",
                    additional_classes="mt-auto mb-auto",
                    hover_class="export-charts",
                ),
            ],
            className="row pt-3 pb-5 charts-filters",
            id="chart-inputs",
        ),
        dcc.Loading(
            html.Div(id="chart-content", style={"height": "calc(100vh - 380px"}),
            type="circle",
        ),
        dcc.Textarea(id="copy-text", style=dict(position="absolute", left="-110%")),
    ]
    if settings.get("github_fork", False):
        body_items.append(
            html.Span(
                html.A(
                    text("Fork me on Github"), href="https://github.com/man-group/dtale"
                ),
                id="forkongithub",
            )
        )
    return body_items
